from pathlib import Path
from unittest.mock import call

import bluesky.plan_stubs as bps
import numpy as np
import pytest
import yaml

from ophyd_async.core import Settings, YamlSettingsProvider, get_mock
from ophyd_async.plan_stubs import (
    apply_settings,
    apply_settings_if_different,
    get_current_settings,
    retrieve_settings,
    store_settings,
)
from ophyd_async.testing import (
    ExampleTable,
    OneOfEverythingDevice,
    ParentOfEverythingDevice,
)

TEST_DATA = Path(__file__).absolute().parent.parent / "test_data"


@pytest.fixture
async def parent_device() -> ParentOfEverythingDevice:
    device = ParentOfEverythingDevice("parent")
    await device.connect(mock=True)
    return device


@pytest.fixture
async def every_parent_device() -> OneOfEverythingDevice:
    device = OneOfEverythingDevice("parent")
    await device.connect(mock=True)
    return device


async def test_get_current_settings(RE, parent_device: ParentOfEverythingDevice):
    expected_values = await parent_device.get_signal_values()

    def my_plan():
        current_settings = yield from get_current_settings(parent_device)
        assert dict(current_settings) == expected_values

    RE(my_plan())


async def test_get_current_config_settings(
    RE, every_parent_device: OneOfEverythingDevice
):
    expected_values = await every_parent_device.get_signal_values()

    def my_plan():
        current_settings = yield from get_current_settings(
            every_parent_device, only_config=True
        )
        current_settings = dict(current_settings)
        for key, value in current_settings.items():
            if isinstance(value, np.ndarray):
                assert np.array_equal(value, expected_values[key])
            else:
                assert value == expected_values[key]

    RE(my_plan())


async def test_store_settings(RE, parent_device: ParentOfEverythingDevice, tmp_path):
    provider = YamlSettingsProvider(tmp_path)

    def my_plan():
        yield from store_settings(provider, "test_file", parent_device)
        with open(tmp_path / "test_file.yaml") as actual_file:
            with open(TEST_DATA / "test_yaml_save.yaml") as expected_file:
                assert yaml.safe_load(actual_file) == yaml.safe_load(expected_file)

    RE(my_plan())


async def test_store_config_settings(
    RE, every_parent_device: OneOfEverythingDevice, tmp_path
):
    provider = YamlSettingsProvider(tmp_path)

    def my_plan():
        yield from store_settings(
            provider, "test_file", every_parent_device, only_config=True
        )
        with open(tmp_path / "test_file.yaml") as actual_file:
            actual_data = yaml.safe_load(actual_file)
        with open(TEST_DATA / "test_yaml_config_save.yaml") as expected_file:
            expected_data = yaml.safe_load(expected_file)
        assert actual_data == expected_data

    RE(my_plan())


async def test_retrieve_and_apply_settings(RE, parent_device: ParentOfEverythingDevice):
    provider = YamlSettingsProvider(TEST_DATA)
    expected_values = await parent_device.get_signal_values()
    serialized_values = {}
    # Override the table to be the serialized version so it compares equal
    for sig, value in expected_values.items():
        if isinstance(value, ExampleTable):
            serialized_values[sig] = {
                k: pytest.approx(v) for k, v in value.model_dump().items()
            }
        else:
            serialized_values[sig] = pytest.approx(value)

    def my_plan():
        m = get_mock(parent_device)
        assert not m.mock_calls
        settings = yield from retrieve_settings(
            provider, "test_yaml_save", parent_device
        )
        assert dict(settings) == serialized_values
        assert not m.mock_calls
        yield from apply_settings(settings)
        assert len(m.mock_calls) == 62
        m.reset_mock()
        assert not m.mock_calls
        yield from apply_settings_if_different(settings, apply_settings)
        assert not m.mock_calls
        yield from bps.abs_set(parent_device.sig_rw, "foo", wait=True)
        assert m.mock_calls == [call.sig_rw.put("foo")]
        m.reset_mock()
        yield from apply_settings_if_different(settings, apply_settings)
        assert m.mock_calls == [call.sig_rw.put("Top level SignalRW")]

    RE(my_plan())


async def test_retrieve_and_apply_config_settings(
    RE, every_parent_device: OneOfEverythingDevice
):
    provider = YamlSettingsProvider(TEST_DATA)
    expected_values = await every_parent_device.get_signal_values()
    serialized_values = {}
    # Override the table to be the serialized version so it compares equal
    for sig, value in expected_values.items():
        if isinstance(value, ExampleTable):
            serialized_values[sig] = {
                k: pytest.approx(v) for k, v in value.model_dump().items()
            }
        else:
            serialized_values[sig] = pytest.approx(value)

    def my_plan():
        m = get_mock(every_parent_device)
        settings = yield from retrieve_settings(
            provider, "test_yaml_config_save", every_parent_device, only_config=True
        )
        assert dict(settings) == serialized_values
        assert not m.mock_calls
        yield from apply_settings(settings)
        assert len(m.mock_calls) == 20
        m.reset_mock()
        assert not m.mock_calls
        yield from apply_settings_if_different(settings, apply_settings)
        assert not m.mock_calls
        yield from bps.abs_set(every_parent_device.a_str, "foo", wait=True)
        assert m.mock_calls == [call.a_str.put("foo")]
        m.reset_mock()
        yield from apply_settings_if_different(settings, apply_settings)
        assert m.mock_calls == [call.a_str.put("test_string")]

    RE(my_plan())


async def test_ignored_settings(RE, parent_device: ParentOfEverythingDevice):
    def my_plan():
        m = get_mock(parent_device)
        settings = Settings(
            parent_device, {parent_device.sig_rw: "foo", parent_device._sig_rw: None}
        )
        yield from apply_settings(settings)
        assert m.mock_calls == [call.sig_rw.put("foo")]

    RE(my_plan())



"""Support for the ADMerlin areaDetector driver.

https://github.com/areaDetector/ADMerlin.
"""

from collections.abc import Sequence
from typing import Annotated as A

from ophyd_async.core import (
    DetectorTriggerLogic,
    PathProvider,
    SignalDict,
    SignalR,
    SignalRW,
    StrictEnum,
)
from ophyd_async.epics.adcore import (
    ADArmLogic,
    ADBaseIO,
    ADWriterType,
    AreaDetector,
    NDPluginBaseIO,
    prepare_exposures,
)
from ophyd_async.epics.core import PvSuffix

__all__ = [
    "MerlinDetector",
    "MerlinDriverIO",
    "MerlinTriggerLogic",
    "MerlinTriggerMode",
]

_MIN_DEAD_TIME = 0.002


class MerlinTriggerMode(StrictEnum):
    """Trigger modes for ADMerlin detector."""

    INTERNAL = "Internal"
    TRIGGER_ENABLE = "Trigger Enable"
    TRIGGER_START_RISING = "Trigger start rising"
    TRIGGER_START_FALLING = "Trigger start falling"
    TRIGGER_BOTH_RISING = "Trigger both rising"
    LVDS_TRIG_ENABLE = "LVDS Trig Enable"
    LVDS_TRIG_START_RISING = "LVDS Trig start rising"
    LVDS_TRIG_START_FALLING = "LVDS Trig start falling"
    LVDS_TRIG_BOTH_RISING = "LVDS Trig both rising"
    SOFTWARE = "Software"


class MerlinDriverIO(ADBaseIO):
    """Driver for merlin model:DU897_BV as deployed on p99.

    This mirrors the interface provided by ADMerlin/db/merlin.template.
    https://github.com/areaDetector/ADMerlin/blob/master/merlinApp/Db/merlin.template
    """

    trigger_mode: A[SignalRW[MerlinTriggerMode], PvSuffix.rbv("TriggerMode")]


# The deadtime of an Merlin controller varies depending on the exact model of camera.
# Ideally we would maximize performance by dynamically retrieving the deadtime at
# runtime. See https://github.com/bluesky/ophyd-async/issues/308
class MerlinTriggerLogic(DetectorTriggerLogic):
    """Trigger logic for MerlinDriverIO."""

    def __init__(self, driver: MerlinDriverIO):
        self.driver = driver

    def get_deadtime(self, config_values: SignalDict) -> float:
        return _MIN_DEAD_TIME

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await self.driver.trigger_mode.set(MerlinTriggerMode.INTERNAL)
        await prepare_exposures(self.driver, num, livetime, deadtime)

    async def prepare_edge(self, num: int, livetime: float):
        # Is this the right trigger mode?
        await self.driver.trigger_mode.set(MerlinTriggerMode.TRIGGER_START_RISING)
        await prepare_exposures(self.driver, num, livetime)


class MerlinDetector(AreaDetector[MerlinDriverIO]):
    """Create an ADMerlin AreaDetector instance.

    :param prefix: EPICS PV prefix for the detector
    :param path_provider: Provider for file paths during acquisition
    :param driver_suffix: Suffix for the driver PV, defaults to "cam1:"
    :param writer_type: Type of file writer (HDF or TIFF)
    :param writer_suffix: Suffix for the writer PV
    :param plugins: Additional areaDetector plugins to include
    :param config_sigs: Additional signals to include in configuration
    :param name: Name for the detector device
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        driver_suffix="cam1:",
        writer_type: ADWriterType | None = ADWriterType.HDF,
        writer_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        driver = MerlinDriverIO(prefix + driver_suffix)
        super().__init__(
            prefix=prefix,
            driver=driver,
            arm_logic=ADArmLogic(driver),
            trigger_logic=MerlinTriggerLogic(driver),
            path_provider=path_provider,
            writer_type=writer_type,
            writer_suffix=writer_suffix,
            plugins=plugins,
            config_sigs=config_sigs,
            name=name,
        )



from __future__ import annotations

from ophyd_async.core import PathProvider, StandardDetector
from ophyd_async.fastcs.core import fastcs_connector

from ._arm_logic import PandaArmLogic
from ._block import CommonPandaBlocks
from ._data_logic import PandaHDFDataLogic
from ._trigger_logic import PandaTriggerLogic

MINIMUM_PANDA_IOC = "0.11.4"


class HDFPanda(CommonPandaBlocks, StandardDetector):
    """PandA with common blocks for standard HDF writing."""

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        name: str = "",
    ):
        error_hint = f"Is PandABlocks-ioc at least version {MINIMUM_PANDA_IOC}?"
        # This has to be first so we make self.pcap
        connector = fastcs_connector(prefix, self, error_hint)
        self.add_detector_logics(
            PandaTriggerLogic(self.pcap),
            PandaArmLogic(self.pcap),
            PandaHDFDataLogic(path_provider, self.data),
        )
        super().__init__(name=name, connector=connector)



from unittest.mock import call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adandor, adcore
from ophyd_async.testing import assert_has_calls


@pytest.fixture
async def test_adandor(
    static_path_provider: StaticPathProvider,
) -> adandor.AndorDetector:
    async with init_devices(mock=True):
        detector = adandor.AndorDetector("PREFIX:", static_path_provider)
    writer = detector.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(writer.file_path_exists, True)
    return detector


def test_pvs_correct(test_adandor: adandor.AndorDetector):
    assert test_adandor.driver.acquire.source == "mock+ca://PREFIX:cam1:Acquire_RBV"
    assert (
        test_adandor.driver.andor_accumulate_period.source
        == "mock+ca://PREFIX:cam1:AndorAccumulatePeriod_RBV"
    )


async def test_deadtime(
    test_adandor: adandor.AndorDetector,
):
    trigger_modes, deadtime = await test_adandor.get_trigger_deadtime()
    assert trigger_modes == {DetectorTrigger.INTERNAL, DetectorTrigger.EXTERNAL_EDGE}
    assert deadtime == 0.1


async def test_prepare_external_edge(
    test_adandor: adandor.AndorDetector,
):
    await test_adandor.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_EDGE,
            number_of_events=5,
            livetime=0.5,
        )
    )
    assert_has_calls(
        test_adandor.driver,
        [
            call.trigger_mode.put(adandor.Andor2TriggerMode.EXT_TRIGGER),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(5),
            call.acquire_time.put(0.5),
            call.acquire.put(True),
        ],
    )


async def test_prepare_forever(
    test_adandor: adandor.AndorDetector,
):
    await test_adandor.prepare(TriggerInfo(number_of_events=0))
    assert_has_calls(
        test_adandor.driver,
        [
            call.trigger_mode.put(adandor.Andor2TriggerMode.INTERNAL),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(999_999),
        ],
    )



"""EPICS Signals over CA or PVA."""

from __future__ import annotations

from collections.abc import Callable
from enum import Enum

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    SignalBackend,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    get_unique,
)

from ._util import EpicsOptions, EpicsSignalBackend, get_pv_basename_and_field


class EpicsProtocol(Enum):
    CA = "ca"
    PVA = "pva"


_default_epics_protocol = EpicsProtocol.CA


def _make_unavailable_function(protocol: str, error: Exception):
    def transport_not_available(*args, **kwargs):
        msg = (
            f"Protocol {protocol} not available, "
            f"did you `pip install ophyd_async[{protocol}]`?"
        )
        raise NotImplementedError(msg) from error

    return transport_not_available


def _make_unavailable_class(
    protocol: str, error: Exception
) -> type[EpicsSignalBackend]:
    class TransportNotAvailable(EpicsSignalBackend):
        __init__ = _make_unavailable_function(protocol, error)

    return TransportNotAvailable


try:
    from ._p4p import PvaSignalBackend, pvget_with_timeout
except ImportError as pva_error:
    PvaSignalBackend = _make_unavailable_class("pva", pva_error)
    pvget_with_timeout = _make_unavailable_function("pva", pva_error)
else:
    _default_epics_protocol = EpicsProtocol.PVA

try:
    from ._aioca import CaSignalBackend
except ImportError as ca_error:
    CaSignalBackend = _make_unavailable_class("ca", ca_error)
else:
    _default_epics_protocol = EpicsProtocol.CA


def split_protocol_from_pv(pv: str) -> tuple[EpicsProtocol, str]:
    split = pv.split("://", 1)
    if len(split) > 1:
        # We got something like pva://mydevice, so use specified comms mode
        scheme, pv = split
        protocol = EpicsProtocol(scheme)
    else:
        # No comms mode specified, use the default
        protocol = _default_epics_protocol
    return protocol, pv


def get_signal_backend_type(protocol: EpicsProtocol) -> type[EpicsSignalBackend]:
    match protocol:
        case EpicsProtocol.CA:
            return CaSignalBackend
        case EpicsProtocol.PVA:
            return PvaSignalBackend
    raise TypeError(f"Unsupported protocol: {protocol}")


def _epics_signal_backend(
    datatype: type[SignalDatatypeT] | None,
    read_pv: str,
    write_pv: str,
    options: EpicsOptions | None = None,
) -> SignalBackend[SignalDatatypeT]:
    """Create an epics signal backend."""
    r_protocol, r_pv = split_protocol_from_pv(read_pv)
    w_protocol, w_pv = split_protocol_from_pv(write_pv)
    protocol = get_unique({read_pv: r_protocol, write_pv: w_protocol}, "protocols")

    signal_backend_type = get_signal_backend_type(protocol)
    return signal_backend_type(datatype, r_pv, w_pv, options)


def epics_signal_rw(
    datatype: type[SignalDatatypeT],
    read_pv: str,
    write_pv: str | None = None,
    name: str = "",
    timeout: float = DEFAULT_TIMEOUT,
    attempts: int = 1,
    wait: bool | Callable[[SignalDatatypeT], bool] = True,
) -> SignalRW[SignalDatatypeT]:
    """Create a `SignalRW` backed by 1 or 2 EPICS PVs.

    :param datatype: Check that the PV is of this type
    :param read_pv: The PV to read and monitor
    :param write_pv: If given, use this PV to write to, otherwise use read_pv
    :param name: The name of the signal (defaults to empty string)
    :param timeout: A timeout to be used when reading (not connecting) this signal
    """
    backend = _epics_signal_backend(
        datatype, read_pv, write_pv or read_pv, EpicsOptions(wait=wait)
    )
    return SignalRW(backend, name=name, timeout=timeout, attempts=attempts)


def epics_signal_rw_rbv(
    datatype: type[SignalDatatypeT],
    write_pv: str,
    read_suffix: str = "_RBV",
    name: str = "",
    timeout: float = DEFAULT_TIMEOUT,
    attempts: int = 1,
    wait: bool | Callable[[SignalDatatypeT], bool] = True,
) -> SignalRW[SignalDatatypeT]:
    """Create a `SignalRW` backed by 1 or 2 EPICS PVs, with a suffix on the readback pv.

    :param datatype: Check that the PV is of this type
    :param write_pv: The PV to write to
    :param read_suffix: Append this suffix to the write pv to create the readback pv
    :param name: The name of the signal (defaults to empty string)
    :param timeout: A timeout to be used when reading (not connecting) this signal
    """
    base_pv, field = get_pv_basename_and_field(write_pv)
    if field is not None:
        read_pv = f"{base_pv}{read_suffix}.{field}"
    else:
        read_pv = f"{write_pv}{read_suffix}"

    return epics_signal_rw(
        datatype, read_pv, write_pv, name, timeout=timeout, attempts=attempts, wait=wait
    )


def epics_signal_r(
    datatype: type[SignalDatatypeT],
    read_pv: str,
    name: str = "",
    timeout: float = DEFAULT_TIMEOUT,
) -> SignalR[SignalDatatypeT]:
    """Create a `SignalR` backed by 1 EPICS PV.

    :param datatype: Check that the PV is of this type
    :param read_pv: The PV to read from
    :param name: The name of the signal (defaults to empty string)
    :param timeout: A timeout to be used when reading (not connecting) this signal
    """
    backend = _epics_signal_backend(datatype, read_pv, read_pv)
    return SignalR(backend, name=name, timeout=timeout)


def epics_signal_w(
    datatype: type[SignalDatatypeT],
    write_pv: str,
    name: str = "",
    timeout: float = DEFAULT_TIMEOUT,
    attempts: int = 1,
    wait: bool | Callable[[SignalDatatypeT], bool] = True,
) -> SignalW[SignalDatatypeT]:
    """Create a `SignalW` backed by 1 EPICS PVs.

    :param datatype: Check that the PV is of this type
    :param write_pv: The PV to write to
    :param name: The name of the signal (defaults to empty string)
    :param timeout: A timeout to be used when reading (not connecting) this signal
    """
    backend = _epics_signal_backend(
        datatype, write_pv, write_pv, EpicsOptions(wait=wait)
    )
    return SignalW(backend, name=name, timeout=timeout, attempts=attempts)


def epics_signal_x(
    write_pv: str, name: str = "", timeout: float = DEFAULT_TIMEOUT
) -> SignalX:
    """Create a `SignalX` backed by 1 EPICS PVs.

    :param write_pv: The PV to write its initial value to on trigger
    :param name: The name of the signal
    :param timeout: A timeout to be used when reading (not connecting) this signal
    """
    backend = _epics_signal_backend(None, write_pv, write_pv)
    return SignalX(backend, name=name, timeout=timeout)



from unittest.mock import call, patch

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adcore, adpilatus
from ophyd_async.testing import assert_has_calls


@pytest.fixture
async def test_adpilatus(
    static_path_provider: StaticPathProvider,
) -> adpilatus.PilatusDetector:
    async with init_devices(mock=True):
        detector = adpilatus.PilatusDetector("PREFIX:", static_path_provider)
    set_mock_value(detector.driver.armed, True)
    writer = detector.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(writer.file_path_exists, True)
    return detector


def test_pvs_correct(test_adpilatus: adpilatus.PilatusDetector):
    assert test_adpilatus.driver.acquire.source == "mock+ca://PREFIX:cam1:Acquire_RBV"
    assert test_adpilatus.driver.armed.source == "mock+ca://PREFIX:cam1:Armed"


@pytest.mark.parametrize(
    "readout_time",
    [adpilatus.PilatusReadoutTime.PILATUS2, adpilatus.PilatusReadoutTime.PILATUS3],
)
async def test_deadtime(readout_time: adpilatus.PilatusReadoutTime, tmp_path):
    path_provider = StaticPathProvider(StaticFilenameProvider("data"), tmp_path)
    pilatus = adpilatus.PilatusDetector("PREFIX:", path_provider, readout_time)
    trigger_modes, deadtime = await pilatus.get_trigger_deadtime()
    assert trigger_modes == {
        DetectorTrigger.INTERNAL,
        DetectorTrigger.EXTERNAL_EDGE,
        DetectorTrigger.EXTERNAL_LEVEL,
    }
    assert deadtime == readout_time


async def test_times_out_if_not_armed(
    test_adpilatus: adpilatus.PilatusDetector,
):
    set_mock_value(test_adpilatus.driver.armed, False)
    with patch(
        "ophyd_async.epics.adcore._arm_logic.DEFAULT_TIMEOUT",
        0.01,
    ):
        with pytest.raises(TimeoutError):
            await test_adpilatus.prepare(
                TriggerInfo(trigger=DetectorTrigger.EXTERNAL_EDGE)
            )


async def test_prepare_external_edge(
    test_adpilatus: adpilatus.PilatusDetector,
):
    await test_adpilatus.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_EDGE,
            number_of_events=5,
            livetime=0.5,
        )
    )
    assert_has_calls(
        test_adpilatus.driver,
        [
            call.trigger_mode.put(adpilatus.PilatusTriggerMode.EXT_TRIGGER),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(5),
            call.acquire_time.put(0.5),
            call.acquire.put(True),
        ],
    )


async def test_prepare_external_level(
    test_adpilatus: adpilatus.PilatusDetector,
):
    await test_adpilatus.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_LEVEL,
            number_of_events=2,
        )
    )
    assert_has_calls(
        test_adpilatus.driver,
        [
            call.trigger_mode.put(adpilatus.PilatusTriggerMode.EXT_ENABLE),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(2),
            call.acquire.put(True),
        ],
    )


async def test_prepare_forever(
    test_adpilatus: adpilatus.PilatusDetector,
):
    await test_adpilatus.prepare(TriggerInfo(number_of_events=0))
    assert_has_calls(
        test_adpilatus.driver,
        [
            call.trigger_mode.put(adpilatus.PilatusTriggerMode.INTERNAL),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(999_999),
        ],
    )



from typing import Annotated as A
from typing import TypeVar

import pytest
from bluesky.protocols import HasHints, Hints

from ophyd_async.core import (
    Device,
    DeviceVector,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    StandardReadable,
    init_devices,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import PviDeviceConnector, SignalDetails


class Block1(Device, HasHints):
    device_vector_signal_x: DeviceVector[SignalX]
    device_vector_signal_rw: DeviceVector[SignalRW[float]]
    signal_x: SignalX
    signal_rw: SignalRW[int]

    @property
    def hints(self) -> Hints:
        return {}


class Block2(Device):
    device_vector: DeviceVector[Block1]
    device: Block1
    signal_x: SignalX
    signal_rw: SignalRW[int]


class Block3(Device):
    device_vector: DeviceVector[Block2]
    device: Block2
    signal_device: Block1
    signal_x: SignalX
    signal_rw: SignalRW[int]


class Block4(StandardReadable):
    device_vector: DeviceVector[Block1]
    device: A[Block1, Format.CHILD]
    signal_x: SignalX
    signal_rw: SignalRW[int]


class Block5(Device):
    signal_rw: SignalRW[int] | None


DeviceT = TypeVar("DeviceT", bound=Device)


def with_pvi_connector(
    device_type: type[DeviceT], prefix: str, name: str = ""
) -> DeviceT:
    connector = PviDeviceConnector(prefix + ":PVI")
    device = device_type(connector=connector, name=name)
    connector.create_children_from_annotations(device)
    return device


async def test_fill_pvi_entries_mock_mode():
    async with init_devices(mock=True):
        test_device = with_pvi_connector(Block3, "PREFIX:")

    # device vectors are typed
    assert isinstance(test_device.device_vector[1], Block2)
    assert isinstance(test_device.device_vector[2], Block2)

    # elements of device vectors are typed recursively
    assert test_device.device_vector[1].signal_rw._connector.backend.datatype is int
    assert isinstance(test_device.device_vector[1].device, Block1)
    assert (
        test_device.device_vector[1].device.signal_rw._connector.backend.datatype is int
    )  # type: ignore
    assert (
        test_device.device_vector[1]
        .device.device_vector_signal_rw[1]
        ._connector.backend.datatype  # type: ignore
        is float
    )

    # top level blocks are typed
    assert isinstance(test_device.signal_device, Block1)
    assert isinstance(test_device.device, Block2)

    # elements of top level blocks are typed recursively
    assert test_device.device.signal_rw._connector.backend.datatype is int  # type: ignore
    assert isinstance(test_device.device.device, Block1)
    assert test_device.device.device.signal_rw._connector.backend.datatype is int  # type: ignore

    assert test_device.signal_rw.parent == test_device
    assert test_device.device_vector.parent == test_device
    assert test_device.device_vector[1].parent == test_device.device_vector
    assert test_device.device_vector[1].device.parent == test_device.device_vector[1]

    assert test_device.name == "test_device"
    assert test_device.device_vector.name == "test_device-device_vector"
    assert test_device.device_vector[1].name == "test_device-device_vector-1"
    assert (
        test_device.device_vector[1].device.name == "test_device-device_vector-1-device"
    )

    # top level signals are typed
    assert test_device.signal_rw._connector.backend.datatype is int


async def test_device_create_children_from_annotations():
    device = with_pvi_connector(Block3, "PREFIX:")

    block_2_device = device.device
    block_1_device = device.device.device
    top_block_1_device = device.signal_device

    # The create_children_from_annotations has made blocks all the way down
    assert isinstance(block_2_device, Block2)
    assert isinstance(block_1_device, Block1)
    assert isinstance(top_block_1_device, Block1)
    assert hasattr(device, "signal_x")
    assert hasattr(device, "signal_rw")
    assert hasattr(top_block_1_device, "signal_rw")

    await device.connect(mock=True)

    # The memory addresses have not changed
    assert device.device is block_2_device
    assert device.device.device is block_1_device
    assert device.signal_device is top_block_1_device


async def test_device_create_device_with_optional_signals():
    device = with_pvi_connector(Block5, "PREFIX:")

    # Makes sure before connecting we create a signal and it exists in the device
    assert hasattr(device, "signal_rw")
    assert isinstance(device.signal_rw, SignalRW)

    await device.connect(mock=True)

    # After connecting if the optional signal is not filled it's set to None
    assert isinstance(device.signal_rw, SignalRW)


async def test_device_create_children_from_annotations_with_device_vectors():
    device = with_pvi_connector(Block4, "PREFIX:", name="test_device")
    await device.connect(mock=True)

    block_1_device = device.device
    assert block_1_device in device._has_hints
    block_2_device_vector = device.device_vector

    assert device.device_vector[1].name == "test_device-device_vector-1"
    assert device.device_vector[2].name == "test_device-device_vector-2"

    # create_children_from_annotiations should have made DeviceVectors
    # and an optional Block, but no signals
    assert hasattr(device, "device_vector")
    assert hasattr(device, "signal_rw")
    assert isinstance(block_2_device_vector, DeviceVector)
    assert isinstance(block_2_device_vector[1], Block1)
    assert len(device.device_vector) == 2
    assert isinstance(block_1_device, Block1)

    # The memory addresses have not changed
    assert device.device is block_1_device
    assert device.device_vector is block_2_device_vector


class NoSignalType(Device):
    a: SignalRW


class NoSignalTypeInVector(Device):
    a: DeviceVector[SignalRW]


@pytest.mark.parametrize("cls", [NoSignalType, NoSignalTypeInVector])
async def test_no_type_annotation_blocks(cls):
    with pytest.raises(TypeError) as exc:
        with_pvi_connector(cls, "PREFIX:")
    assert str(exc.value) == (
        f"{cls.__name__}.a: Expected SignalX or SignalR/W/RW[type], "
        "got <class 'ophyd_async.core._signal.SignalRW'>"
    )


@pytest.mark.parametrize(
    "mock_entry, expected_signal_type",
    [
        ({"r": "read_pv"}, SignalR),
        ({"r": "read_pv", "w": "write_pv"}, SignalRW),
        ({"rw": "read_and_write_pv"}, SignalRW),
        ({"w": "write_pv"}, SignalW),
        ({"x": "triggerable_pv"}, SignalX),
        ({"invalid": "invalid_pv"}, None),
    ],
)
async def test_correctly_setting_signal_type_from_signal_details(
    mock_entry, expected_signal_type
):
    if not expected_signal_type:
        with pytest.raises(TypeError) as exc:
            SignalDetails.from_entry(mock_entry)
        assert "Can't process entry" in str(exc.value)
    else:
        details = SignalDetails.from_entry(mock_entry)
        assert details.signal_type == expected_signal_type



import asyncio

from ophyd_async.core import DetectorTriggerLogic, SignalDict

from ._io import EigerDetectorIO, EigerTriggerMode


async def _prepare_detector(
    detector: EigerDetectorIO,
    trigger_mode: EigerTriggerMode,
    num: int,
    livetime: float = 0.0,
):
    coros = [
        detector.trigger_mode.set(trigger_mode),
        detector.nimages.set(num),
    ]
    if livetime:
        coros += [
            detector.count_time.set(livetime),
            detector.frame_time.set(livetime),
        ]
    await asyncio.gather(*coros)


class EigerTriggerLogic(DetectorTriggerLogic):
    def __init__(
        self,
        detector: EigerDetectorIO,
    ) -> None:
        self.detector = detector

    def get_deadtime(self, config_values: SignalDict) -> float:
        # See https://media.dectris.com/filer_public/30/14/3014704e-5f3b-43ba-8ccf-8ef720e60d2a/240202_usermanual_eiger2.pdf
        return 0.0001

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await _prepare_detector(self.detector, EigerTriggerMode.INTERNAL, num, livetime)

    async def prepare_edge(self, num: int, livetime: float):
        await _prepare_detector(self.detector, EigerTriggerMode.EDGE, num, livetime)

    async def prepare_level(self, num: int):
        await _prepare_detector(self.detector, EigerTriggerMode.GATE, num)



import asyncio
import logging

from ophyd_async.core import DetectorTriggerLogic, SignalDict, SignalRW

from ._io import AcquisitionType, JungfrauDriverIO, JungfrauTriggerMode, PedestalMode

# Deadtime is dependant on a wide combination of settings and on trigger mode
# but this is safe upper-limit
JUNGFRAU_DEADTIME_S = 2e-5

logger = logging.getLogger("ophyd_async")


async def get_and_reset_acquisition_type(
    acquisition_type: SignalRW[AcquisitionType],
) -> AcquisitionType:
    acquisition_type_value = await acquisition_type.get_value()
    # Pedestal mode is one-shot, so set back the request value
    await acquisition_type.set(AcquisitionType.STANDARD)
    logger.info(f"Preparing Jungfrau in {acquisition_type_value} mode.")
    return acquisition_type_value


async def prepare_exposures(
    detector: JungfrauDriverIO, livetime: float, deadtime: float = 0
):
    # ValueErrors and warnings in this function come from the jungfrau operation
    # docs: https://rtd.xfel.eu/docs/jungfrau-detector-documentation/en/latest/operation.html
    if not livetime:
        livetime = await detector.exposure_time.get_value()
    if livetime < 2e-6:
        logger.warning("Exposure time shorter than 2μs is not recommended")
    if not deadtime:
        deadtime = JUNGFRAU_DEADTIME_S
    period_between_frames = livetime + deadtime
    if period_between_frames < JUNGFRAU_DEADTIME_S:
        raise ValueError(
            f"Period between frames (exposure time + deadtime) = "
            f"{period_between_frames}s cannot be lower than minimum detector "
            f"deadtime {JUNGFRAU_DEADTIME_S}"
        )
    await asyncio.gather(
        detector.period_between_frames.set(period_between_frames),
        detector.exposure_time.set(livetime),
    )


async def prepare_standard_mode(
    detector: JungfrauDriverIO,
    trigger_mode: JungfrauTriggerMode,
    num: int,
    livetime: float,
    deadtime: float = 0,
) -> None:
    await asyncio.gather(
        detector.trigger_mode.set(trigger_mode),
        prepare_exposures(detector, livetime, deadtime),
        detector.frames_per_acq.set(num),
    )


async def prepare_pedestal_mode(
    detector: JungfrauDriverIO, num: int, livetime: float, deadtime: float
):
    frames, loops = await asyncio.gather(
        detector.pedestal_mode_frames.get_value(),
        detector.pedestal_mode_loops.get_value(),
    )
    if 2 * frames * loops != num:
        # No. events is double the pedestal loops,
        # since pedestal scan does the entire loop
        # twice.
        raise ValueError(
            f"Invalid trigger info for pedestal mode. "
            f"Number {num} must be equal to 2 * {frames} * {loops}. "
            f"Was create_jungfrau_pedestal_triggering_info used?"
        )
    await asyncio.gather(
        detector.trigger_mode.set(JungfrauTriggerMode.INTERNAL),
        prepare_exposures(detector, livetime, deadtime),
    )
    # Setting signals once the detector is in pedestal mode can cause errors,
    # so do this last
    await detector.pedestal_mode_state.set(PedestalMode.ON)


class JungfrauTriggerLogic(DetectorTriggerLogic):
    def __init__(
        self, detector: JungfrauDriverIO, acquisition_type: SignalRW[AcquisitionType]
    ):
        self.detector = detector
        self.acquisition_type = acquisition_type

    # Deadtime here is really used as "time between frames"
    def get_deadtime(self, config_values: SignalDict) -> float:
        return JUNGFRAU_DEADTIME_S

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        acquisition_type = await get_and_reset_acquisition_type(self.acquisition_type)
        if acquisition_type == AcquisitionType.PEDESTAL:
            await prepare_pedestal_mode(self.detector, num, livetime, deadtime)
        else:
            await prepare_standard_mode(
                self.detector, JungfrauTriggerMode.INTERNAL, num, livetime, deadtime
            )

    async def prepare_edge(self, num: int, livetime: float):
        acquisition_type = await get_and_reset_acquisition_type(self.acquisition_type)
        if acquisition_type == AcquisitionType.PEDESTAL:
            raise ValueError(
                "Jungfrau must be triggered internally while in pedestal mode."
            )
        else:
            await prepare_standard_mode(
                self.detector, JungfrauTriggerMode.EXTERNAL, num, livetime
            )



"""Core components of the areaDetector software.

https://github.com/areaDetector/ADCore
"""

from ._arm_logic import ADArmLogic, ADContAcqArmLogic
from ._data_logic import (
    ADHDFDataLogic,
    ADMultipartDataLogic,
    ADWriterType,
    NDArrayDescription,
    PluginSignalDataLogic,
)
from ._detector import AreaDetector, ContAcqDetector
from ._io import (
    NDROIIO,
    ADBaseColorMode,
    ADBaseDataType,
    ADBaseIO,
    ADCompression,
    ADFileWriteMode,
    ADImageMode,
    ADState,
    NDArrayBaseIO,
    NDCBFlushOnSoftTrgMode,
    NDCircularBuffIO,
    NDFileHDF5IO,
    NDFileIO,
    NDPluginBaseIO,
    NDPluginFileIO,
    NDROIStatIO,
    NDROIStatNIO,
    NDStatsIO,
)
from ._ndattribute import (
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    NDAttributePvDbrType,
    ndattributes_to_xml,
)
from ._plan_stubs import setup_ndattributes, setup_ndstats_sum
from ._trigger_logic import ADContAcqTriggerLogic, prepare_exposures

__all__ = [
    # ADCore IOs
    "ADBaseDataType",
    "ADBaseColorMode",
    "NDArrayBaseIO",
    "ADImageMode",
    "ADState",
    "ADBaseIO",
    "NDPluginBaseIO",
    "NDROIIO",
    "NDStatsIO",
    "NDROIStatNIO",
    "NDROIStatIO",
    "NDCBFlushOnSoftTrgMode",
    "NDCircularBuffIO",
    "ADFileWriteMode",
    "NDFileIO",
    "NDPluginFileIO",
    "ADCompression",
    "NDFileHDF5IO",
    # TriggerLogic
    "prepare_exposures",
    "ADContAcqTriggerLogic",
    # ArmLogic
    "ADArmLogic",
    "ADContAcqArmLogic",
    # DataLogic
    "NDArrayDescription",
    "PluginSignalDataLogic",
    "ADHDFDataLogic",
    "ADMultipartDataLogic",
    "ADWriterType",
    # Detector
    "AreaDetector",
    "ContAcqDetector",
    # NDAttributes
    "NDAttributeDataType",
    "NDAttributePvDbrType",
    "NDAttributePv",
    "NDAttributeParam",
    "ndattributes_to_xml",
    "setup_ndattributes",
    "setup_ndstats_sum",
]



from ophyd_async.core import AsyncReadable, StandardFlyer
from ophyd_async.sim import SimBlobDetector, SimMotor


async def test_readable():
    assert isinstance(SimMotor, AsyncReadable)
    assert isinstance(SimBlobDetector, AsyncReadable)
    assert not isinstance(StandardFlyer, AsyncReadable)



from typing import Annotated as A

from ophyd_async.core import SignalR, SignalRW, StandardReadable, StrictEnum
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import EpicsDevice, PvSuffix


class EnergyMode(StrictEnum):
    """Energy mode for `DemoPointDetectorChannel`."""

    LOW = "Low Energy"
    """Low energy mode"""

    HIGH = "High Energy"
    """High energy mode"""


class DemoPointDetectorChannel(StandardReadable, EpicsDevice):
    """A channel for `DemoPointDetector` with int value based on X and Y Motors."""

    value: A[SignalR[int], PvSuffix("Value"), Format.HINTED_UNCACHED_SIGNAL]
    mode: A[SignalRW[EnergyMode], PvSuffix("Mode"), Format.CONFIG_SIGNAL]



"""Module which defines abstract classes to work with detectors."""

import asyncio
import functools
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable, Callable, Sequence
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from typing import cast

from bluesky.protocols import (
    Collectable,
    Flyable,
    HasHints,
    Hints,
    Preparable,
    Reading,
    Stageable,
    StreamAsset,
    Triggerable,
    WritesStreamAssets,
)
from event_model import DataKey
from pydantic import Field, NonNegativeInt, PositiveInt, computed_field

from ._data_providers import ReadableDataProvider, StreamableDataProvider
from ._device import Device
from ._protocol import AsyncConfigurable, AsyncReadable
from ._settings import Settings
from ._signal import (
    SignalDict,
    SignalR,
    SignalRW,
    observe_signals_value,
    soft_signal_rw,
)
from ._status import AsyncStatus, WatchableAsyncStatus
from ._utils import (
    DEFAULT_TIMEOUT,
    ConfinedModel,
    WatcherUpdate,
    error_if_none,
    merge_gathered_dicts,
)


class DetectorTrigger(Enum):
    """Type of mechanism for triggering a detector to take exposures."""

    INTERNAL = "INTERNAL"
    """On arm generate internally timed exposures"""

    EXTERNAL_EDGE = "EXTERNAL_EDGE"
    """On every (normally rising) edge of an external input generate an internally
    timed exposure"""

    EXTERNAL_LEVEL = "EXTERNAL_LEVEL"
    """On a rising edge of an external input start an exposure, ending on the falling
    edge"""


class TriggerInfo(ConfinedModel):
    """Information required to setup `trigger` or `kickoff` on a `StandardDetector`."""

    trigger: DetectorTrigger = Field(default=DetectorTrigger.INTERNAL)
    """What sort of triggering should the detector be set for."""

    livetime: float = Field(default=0.0, ge=0.0)
    """For INTERNAL or EXTERNAL_EDGE triggering, how long should each exposure be.
    0 means whatever is currently set."""

    deadtime: float = Field(default=0.0, ge=0.0)
    """For INTERNAL triggering, how long should be left between each exposure.
    0 means use the minimum the detector supports."""

    exposures_per_collection: PositiveInt = Field(default=1)
    """An exposure corresponds to a single trigger sent to the detector.
    If many exposures are averaged together on the detector or in a processing
    chain to make a single collection that is exposed to bluesky as data then
    this number should be set to the number of exposures to be processed into a
    single collection."""

    collections_per_event: PositiveInt = Field(default=1)
    """A collection is exposed to bluesky as data, but different detectors can
    be set to have a different number of collections per event so that multiple
    collections from a faster detector can be zipped with a single collection
    from a slower detector. E.g. if number_of_events=10 and
    collections_per_event=5 then the detector will take 50 exposures, but
    publish 10 StreamDatum indices, and describe() will show a shape of (5, h,
    w) for each.
    """

    number_of_events: NonNegativeInt = Field(default=1)
    """Number of bluesky events that will be emitted, (0 means infinite)."""

    exposure_timeout: float = Field(
        default_factory=lambda d: d["livetime"] + d["deadtime"] + DEFAULT_TIMEOUT,
        gt=0,
    )
    """What is the maximum timeout on waiting for an exposure"""

    @computed_field
    @cached_property
    def number_of_collections(self) -> int:
        return self.number_of_events * self.collections_per_event

    @computed_field
    @cached_property
    def number_of_exposures(self) -> int:
        return self.number_of_collections * self.exposures_per_collection


class DetectorTriggerLogic:
    """Logic for configuring detector triggering modes.

    This class defines the interface for detector trigger configuration, handling
    both internal and external triggering modes. Implementations should provide
    detector-specific logic for preparing the detector to operate in different
    trigger modes and manage exposure parameters.

    The class manages:
    - Configuration signals that should appear in detector metadata
    - Deadtime calculations based on detector configuration
    - Preparation for internal (self-triggered) exposures
    - Preparation for external edge-triggered exposures
    - Preparation for external level-triggered exposures
    - Multi-exposure collection batching

    Subclasses must implement the appropriate `prepare_*` method for any trigger
    mode the detector supports, `get_deadtime` if it supports external
    triggering, and `config_sigs` if the deadtime would vary according to
    detector parameters.
    """

    def config_sigs(self) -> set[SignalR]:
        """Return the signals that should appear in read_configuration."""
        return set()

    def get_deadtime(self, config_values: SignalDict) -> float:
        """Return the deadtime in seconds for the detector.

        :param config_values: the value of each signal in `config_sigs`
        """
        raise NotImplementedError(self)

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        """Prepare the detector to take internally triggered exposures.

        :param num: the number of exposures to take
        :param livetime: how long the exposure should be, 0 means what is currently set
        :param deadtime: how long between exposures, 0 means the shortest possible
        """
        raise NotImplementedError(self)

    async def prepare_edge(self, num: int, livetime: float):
        """Prepare the detector to take external edge triggered exposures.

        :param num: the number of exposures to take
        :param livetime: how long the exposure should be, 0 means what is currently set
        """
        raise NotImplementedError(self)

    async def prepare_level(self, num: int):
        """Prepare the detector to take external level triggered exposures.

        :param num: the number of exposures to take
        """
        raise NotImplementedError(self)

    async def prepare_exposures_per_collection(self, exposures_per_collection: int):
        """Prepare processing of multiple exposures into a single collection.

        :param exposures_per_collection:
            number of exposures to process into each collection
        """
        raise NotImplementedError(self)


def _logic_supported(base_class, method) -> bool:
    # If the function that is bound in a subclass is the same as the function
    # attached to the superclass, then the subclass has not overridden it, so
    # this method is not supported by the subclass.
    return method.__func__ is not getattr(base_class, method.__name__)


_trigger_logic_supported = functools.partial(_logic_supported, DetectorTriggerLogic)


def _get_supported_triggers(
    trigger_logic: DetectorTriggerLogic,
) -> set[DetectorTrigger]:
    supported_triggers = set()
    if _trigger_logic_supported(trigger_logic.prepare_internal):
        supported_triggers.add(DetectorTrigger.INTERNAL)
    if _trigger_logic_supported(trigger_logic.prepare_edge):
        supported_triggers.add(DetectorTrigger.EXTERNAL_EDGE)
    if _trigger_logic_supported(trigger_logic.prepare_level):
        supported_triggers.add(DetectorTrigger.EXTERNAL_LEVEL)
    return supported_triggers


class DetectorArmLogic(ABC):
    """Abstract base class for detector arming and disarming logic.

    Implementations must provide methods to arm the detector, wait for it to become
    idle, and disarm it. This interface allows for detector-specific behavior during
    the arm/disarm lifecycle.
    """

    @abstractmethod
    async def arm(self):
        """Arm the detector, waiting until it is armed."""

    @abstractmethod
    async def wait_for_idle(self):
        """Wait for the detector to be disarmed or idle."""

    @abstractmethod
    async def disarm(self):
        """Disarm the detector, return detector to an idle state."""


def _all_the_same(collections_written: set[int]) -> int:
    """Ensure all collection counts are the same, raising an error if they differ.

    :param collections_written: Set of collection counts from different providers
    :return: The single collection count value
    :raises RuntimeError: If the set contains more than one distinct value
    """
    if len(collections_written) != 1:
        msg = (
            "Detectors have written different numbers of collections: "
            + f"{collections_written}"
        )
        raise RuntimeError(msg)
    return collections_written.pop()


async def _get_collections_written(
    data_providers: Sequence[StreamableDataProvider],
    reducer: Callable[[set[int]], int] = _all_the_same,
) -> int:
    """Return a single collections_written value for the given providers.

    By default this function ensures all providers agree and returns that
    single value. If `reducer` is provided it will be called with the set of
    observed values and should return a single int to use.
    """
    # Work out where all the streamable data providers are up to
    collections_written = set(
        await asyncio.gather(
            *[sdp.collections_written_signal.get_value() for sdp in data_providers]
        )
    )
    if collections_written:
        # Let our reducer decide how to return a single int
        return reducer(collections_written)
    else:
        # There are none, this is valid as we then don't use the value anywhere
        # so just return 0
        return 0


class DetectorDataLogic:
    """Abstract base class for detector data logic and handling.

    Implementations must implement either prepare_unbounded for data sources
    that work with step scans as well as flyscans, or prepare_single for those
    that only work with step scans.
    """

    #: Add this suffix to the detector name to specify the datakey. These need to be
    #: different for each DetectorDataLogic added to a detector
    datakey_suffix: str = ""

    async def prepare_single(self, datakey_name: str) -> ReadableDataProvider:
        """Provider can only work for a single event."""
        raise NotImplementedError(self)

    async def prepare_unbounded(self, datakey_name: str) -> StreamableDataProvider:
        """Provider can work for an unbounded number of collections."""
        raise NotImplementedError(self)

    def get_hinted_fields(self, datakey_name: str) -> Sequence[str]:
        """Return the hinted streams."""
        return []

    async def stop(self) -> None:
        """Stop taking data."""
        return None


_data_logic_supported = functools.partial(_logic_supported, DetectorDataLogic)


@dataclass
class _PrepareCtx:
    trigger_info: TriggerInfo
    readable_data_providers: Sequence[ReadableDataProvider]
    streamable_data_providers: Sequence[StreamableDataProvider]
    collections_written: int


@dataclass
class _KickoffCtx:
    trigger_info: TriggerInfo
    data_providers: Sequence[StreamableDataProvider]
    collections_written: int
    collections_requested: int
    is_last_kickoff: bool


class StandardDetector(
    Device,
    Stageable,
    AsyncConfigurable,
    AsyncReadable,
    Triggerable,
    Preparable,
    Flyable,
    Collectable,
    WritesStreamAssets,
    HasHints,
):
    """Detector base class for step and fly scanning detectors.

    Aggregates trigger, arm, reading or stream logic together.
    """

    # Logic for the detector
    _trigger_logic: DetectorTriggerLogic | None = None
    _arm_logic: DetectorArmLogic | None = None
    _data_logics: Sequence[DetectorDataLogic] = ()
    # Signals to include in read_configuration
    _config_signals: Sequence[SignalR] = ()
    # Context produced by prepare, used by trigger and kickoff
    _prepare_ctx: _PrepareCtx | None = None
    # Context produced by kickoff, used by complete
    _kickoff_ctx: _KickoffCtx | None = None
    # The triggers that are supported by the trigger logic
    _supported_triggers: set[DetectorTrigger] = {DetectorTrigger.INTERNAL}

    # Report the number of events for the next kickoff
    @cached_property
    def events_to_kickoff(self) -> SignalRW[int]:
        # TODO: only allow this to be revised down when trigger_info.number_of_events >1
        # and we have a reusable data provider
        # requries https://github.com/bluesky/ophyd-async/issues/1119
        signal = soft_signal_rw(int)
        # Name and parent this manually as `Device` doesn't know how to deal with cached
        # properties
        signal.parent = self
        signal.set_name(f"{self.name}-events_to_kickoff")
        return signal

    def add_detector_logics(
        self, *logics: DetectorTriggerLogic | DetectorArmLogic | DetectorDataLogic
    ) -> None:
        """Add arm, trigger or data logic to the detector.

        :param logic: The logic to add
        """
        for logic in logics:
            if isinstance(logic, DetectorTriggerLogic):
                if self._trigger_logic is not None:
                    raise RuntimeError("Detector already has trigger logic")
                self._trigger_logic = logic
                # Store the triggers that are supported
                self._supported_triggers = _get_supported_triggers(logic)
                # Add the config signals it needs
                self.add_config_signals(*logic.config_sigs())
            elif isinstance(logic, DetectorArmLogic):
                if self._arm_logic is not None:
                    raise RuntimeError("Detector already has arm logic")
                self._arm_logic = logic
            elif isinstance(logic, DetectorDataLogic):
                self._data_logics = (*self._data_logics, logic)
            else:
                raise TypeError(f"Unknown logic type: {type(logic)}")

    def add_config_signals(self, *signals: SignalR) -> None:
        """Add a signal to read_configuration().

        :param sig: The signal to add
        """
        self._config_signals = (*self._config_signals, *signals)

    async def _disarm_and_stop(self):
        coros = [data_logic.stop() for data_logic in self._data_logics]
        if self._arm_logic:
            coros.append(self._arm_logic.disarm())
        await asyncio.gather(*coros)

    async def get_trigger_deadtime(
        self, settings: Settings | None = None
    ) -> tuple[set[DetectorTrigger], float | None]:
        """Get supported trigger types and deadtime for the detector.

        :param settings: Optional settings to use when getting configuration values
        :return: Tuple of supported trigger types and deadtime in seconds
        """
        if self._trigger_logic and _trigger_logic_supported(
            self._trigger_logic.get_deadtime
        ):
            config_values = SignalDict()
            for sig in self._trigger_logic.config_sigs():
                if settings and sig in settings:
                    # Use value from settings if it is in there
                    # cast to a SignalRW because settings can only contain those
                    config_values[sig] = settings[cast(SignalRW, sig)]
                else:
                    # Get the value live
                    config_values[sig] = await sig.get_value()
            deadtime = self._trigger_logic.get_deadtime(config_values)
        else:
            deadtime = None
        return self._supported_triggers, deadtime

    @AsyncStatus.wrap
    async def stage(self) -> None:
        """Make sure the detector is idle and ready to be used."""
        await self._disarm_and_stop()
        self._prepare_ctx = None
        self._kickoff_ctx = None
        await self.events_to_kickoff.set(0)

    async def _update_prepare_context(self, trigger_info: TriggerInfo) -> None:
        # The only thing that would stop us being able to reuse a provider is
        # if the collections_per_event changes, as that would change the
        # StreamResource shape. All other TriggerInfo parameters (exposures, livetime,
        # etc.) don't affect the data provider configuration.
        if (
            self._prepare_ctx
            and self._prepare_ctx.trigger_info.collections_per_event
            == trigger_info.collections_per_event
        ):
            # Reuse the existing data providers
            readable_data_providers = self._prepare_ctx.readable_data_providers
            streamable_data_providers = self._prepare_ctx.streamable_data_providers
        else:
            # Stop the existing providers if there is a context and make new ones
            if self._prepare_ctx:
                for data_logic in self._data_logics:
                    await data_logic.stop()
            # Setup the data logic for the right number of collections
            streamable_coros: list[Awaitable[StreamableDataProvider]] = []
            readable_coros: list[Awaitable[ReadableDataProvider]] = []
            for dl in self._data_logics:
                if _data_logic_supported(dl.prepare_unbounded):
                    streamable_coros.append(
                        dl.prepare_unbounded(self.name + dl.datakey_suffix)
                    )
                elif _data_logic_supported(dl.prepare_single):
                    if trigger_info.number_of_collections > 1:
                        raise RuntimeError(
                            f"Multiple collections not supported by"
                            f" {self.name + dl.datakey_suffix}"
                        )
                    readable_coros.append(
                        dl.prepare_single(self.name + dl.datakey_suffix)
                    )
                else:
                    msg = f"DataLogic hasn't overridden any prepare_* methods {dl}"
                    raise RuntimeError(msg)
            streamable_data_providers, readable_data_providers = await asyncio.gather(
                asyncio.gather(*streamable_coros),
                asyncio.gather(*readable_coros),
            )
        # Stash the prepare context so we can use it in trigger/kickoff
        self._prepare_ctx = _PrepareCtx(
            trigger_info=trigger_info,
            streamable_data_providers=streamable_data_providers,
            readable_data_providers=readable_data_providers,
            collections_written=await _get_collections_written(
                streamable_data_providers
            ),
        )

    async def _wait_for_index(
        self,
        data_providers: Sequence[StreamableDataProvider],
        trigger_info: TriggerInfo,
        initial_collections_written: int,
        collections_requested: int,
        wait_for_idle: bool,
    ) -> AsyncIterator[WatcherUpdate]:
        start_time = time.monotonic()
        current_collections_written = {
            dp.collections_written_signal: initial_collections_written
            for dp in data_providers
        }
        collections_per_event = trigger_info.collections_per_event
        target_collections_written = initial_collections_written + collections_requested
        if data_providers:
            async for sig, value in observe_signals_value(
                *current_collections_written.keys(),
                timeout=trigger_info.exposure_timeout,
            ):
                current_collections_written[sig] = value
                collections_written = min(current_collections_written.values())
                yield WatcherUpdate(
                    name=self.name,
                    current=collections_written // collections_per_event,
                    initial=initial_collections_written // collections_per_event,
                    target=target_collections_written // collections_per_event,
                    unit="",
                    precision=0,
                    time_elapsed=time.monotonic() - start_time,
                )
                if collections_written >= target_collections_written:
                    break
        if self._arm_logic and wait_for_idle:
            await self._arm_logic.wait_for_idle()

    @AsyncStatus.wrap
    async def prepare(self, value: TriggerInfo) -> None:
        """Prepare the detector for a number of triggers.

        :param value: TriggerInfo describing how to trigger the detector
        """
        if self._trigger_logic and _trigger_logic_supported(
            self._trigger_logic.prepare_exposures_per_collection
        ):
            # If we can do multiple exposures per collection then set it up
            # even if there was only 1 requested to clear previous settings
            await self._trigger_logic.prepare_exposures_per_collection(
                value.exposures_per_collection
            )
        elif value.exposures_per_collection != 1:
            raise ValueError(
                f"Multiple exposures per collection not supported by {self}"
            )
        # Setup the trigger logic for the right number of exposures
        if value.trigger not in self._supported_triggers:
            format_triggers = ", ".join(
                sorted(t.name for t in self._supported_triggers)
            )
            raise ValueError(
                f"Trigger type {value.trigger} not supported by '{self.name}', "
                f"supported types are: [{format_triggers}]"
            )
        if self._trigger_logic:
            match value.trigger:
                case DetectorTrigger.INTERNAL:
                    await self._trigger_logic.prepare_internal(
                        num=value.number_of_exposures,
                        livetime=value.livetime,
                        deadtime=value.deadtime,
                    )
                case DetectorTrigger.EXTERNAL_EDGE:
                    await self._trigger_logic.prepare_edge(
                        num=value.number_of_exposures,
                        livetime=value.livetime,
                    )
                case DetectorTrigger.EXTERNAL_LEVEL:
                    await self._trigger_logic.prepare_level(
                        num=value.number_of_exposures,
                    )
        elif value.livetime != 0.0 or value.deadtime != 0.0:
            raise ValueError(
                f"Detector {self.name} has no trigger logic, so cannot set livetime or "
                "deadtime"
            )
        # NOTE: this section must come after preparing the trigger logic as we may
        # use parameters from it to determine datatype for the streams
        await self._update_prepare_context(value)
        # Tell people how many collections we will acquire for
        await self.events_to_kickoff.set(value.number_of_events)
        # External triggering can arm now
        if self._arm_logic and value.trigger != DetectorTrigger.INTERNAL:
            await self._arm_logic.arm()

    @WatchableAsyncStatus.wrap
    async def trigger(self) -> AsyncIterator[WatcherUpdate[int]]:
        if self._prepare_ctx is None:
            # If a prepare has not been done since stage, do an implicit one here
            await self.prepare(TriggerInfo())
        else:
            # Check the one that was provided is suitable for triggering
            trigger_info = self._prepare_ctx.trigger_info
            if trigger_info.number_of_events != 1:
                msg = (
                    "trigger() is not supported for multiple events, the detector was "
                    f"prepared with number_of_events={trigger_info.number_of_events}."
                )
                raise ValueError(msg)
            # Ensure the data provider is still usable
            await self._update_prepare_context(trigger_info)
        ctx = error_if_none(self._prepare_ctx, "Prepare should have been run")
        # Arm the detector and wait for it to finish.
        if self._arm_logic:
            await self._arm_logic.arm()
        async for update in self._wait_for_index(
            data_providers=ctx.streamable_data_providers,
            trigger_info=ctx.trigger_info,
            initial_collections_written=ctx.collections_written,
            collections_requested=1,
            wait_for_idle=True,
        ):
            yield update

    @AsyncStatus.wrap
    async def kickoff(self):
        ctx = error_if_none(self._prepare_ctx, "Prepare not called")
        if not ctx.streamable_data_providers:
            raise ValueError(
                f"Detector {self.name} is not streamable, so cannot kickoff"
            )
        collections_written, events_to_kickoff = await asyncio.gather(
            _get_collections_written(ctx.streamable_data_providers),
            self.events_to_kickoff.get_value(),
        )
        collections_requested = (
            events_to_kickoff * ctx.trigger_info.collections_per_event
        )
        last_requested_collection = collections_written + collections_requested
        last_expected_collection = (
            ctx.collections_written + ctx.trigger_info.number_of_collections
        )
        if last_requested_collection > last_expected_collection:
            msg = (
                f"Kickoff requested {collections_written}:{last_requested_collection}, "
                f"but detector was only prepared up to {last_expected_collection}"
            )
            raise RuntimeError(msg)
        self._kickoff_ctx = _KickoffCtx(
            trigger_info=ctx.trigger_info,
            data_providers=ctx.streamable_data_providers,
            collections_written=collections_written,
            collections_requested=collections_requested,
            is_last_kickoff=last_requested_collection == last_expected_collection,
        )
        # External trigering has been armed already, internal should arm now
        if self._arm_logic and ctx.trigger_info.trigger == DetectorTrigger.INTERNAL:
            await self._arm_logic.arm()

    @WatchableAsyncStatus.wrap
    async def complete(self):
        ctx = error_if_none(self._kickoff_ctx, "Kickoff not called")
        async for update in self._wait_for_index(
            data_providers=ctx.data_providers,
            trigger_info=ctx.trigger_info,
            initial_collections_written=ctx.collections_written,
            collections_requested=ctx.collections_requested,
            wait_for_idle=ctx.is_last_kickoff,
        ):
            yield update

    async def describe_configuration(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts(
            sig.describe() for sig in self._config_signals
        )

    async def read_configuration(self) -> dict[str, Reading]:
        return await merge_gathered_dicts(sig.read() for sig in self._config_signals)

    async def describe(self) -> dict[str, DataKey]:
        ctx = error_if_none(self._prepare_ctx, "Prepare not run")
        # Readable and Streamable data providers produce data during read
        coros = [dp.make_datakeys() for dp in ctx.readable_data_providers] + [
            dp.make_datakeys(ctx.trigger_info.collections_per_event)
            for dp in ctx.streamable_data_providers
        ]
        return await merge_gathered_dicts(coros)

    async def describe_collect(self) -> dict[str, DataKey]:
        ctx = error_if_none(self._prepare_ctx, "Prepare not run")
        # Only streamable data providers produce data during collect
        coros = [
            dp.make_datakeys(ctx.trigger_info.collections_per_event)
            for dp in ctx.streamable_data_providers
        ]
        return await merge_gathered_dicts(coros)

    @property
    def hints(self) -> Hints:
        fields: list[str] = []
        for dl in self._data_logics:
            fields.extend(dl.get_hinted_fields(self.name + dl.datakey_suffix))
        return Hints(fields=fields)

    async def read(self) -> dict[str, Reading]:
        ctx = error_if_none(self._prepare_ctx, "Prepare not called")
        return await merge_gathered_dicts(
            dp.make_readings() for dp in ctx.readable_data_providers
        )

    async def collect_asset_docs(
        self, index: int | None = None
    ) -> AsyncIterator[StreamAsset]:
        # Collect stream datum documents for all indices written.
        ctx = error_if_none(self._prepare_ctx, "Prepare not called")
        if index is None:
            # The index is optional, and provided for fly scans, if there is
            # more than one detector to make sure they collect in step
            index = await self.get_index()
        for data_provider in ctx.streamable_data_providers:
            async for doc in data_provider.make_stream_docs(
                collections_written=index * ctx.trigger_info.collections_per_event,
                collections_per_event=ctx.trigger_info.collections_per_event,
            ):
                yield doc

    async def get_index(self) -> int:
        ctx = error_if_none(self._prepare_ctx, "Prepare not called")
        min_collections_written = await _get_collections_written(
            ctx.streamable_data_providers, reducer=min
        )
        return min_collections_written // ctx.trigger_info.collections_per_event

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        """Disarm the detector and stop file writing."""
        await self._disarm_and_stop()



from ophyd_async.core import DetectorTriggerLogic

from ._pattern_generator import PatternGenerator


class BlobTriggerLogic(DetectorTriggerLogic):
    def __init__(self, pattern_generator: PatternGenerator):
        self.pattern_generator = pattern_generator

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        self.pattern_generator.setup_acquisition_parameters(
            exposure=livetime,
            period=livetime + deadtime,
            number_of_frames=num,
        )



import asyncio
from collections.abc import Mapping, Sequence
from contextlib import AbstractContextManager
from typing import Any, cast
from unittest.mock import Mock, call

import pytest
from bluesky.protocols import Reading
from event_model import DataKey

from ophyd_async.core import (
    AsyncConfigurable,
    AsyncReadable,
    Device,
    Signal,
    SignalDatatypeT,
    SignalR,
    Table,
    WatchableAsyncStatus,
    Watcher,
    get_mock,
)

from ._utils import T


def partial_reading(val: Any) -> Mapping[str, Any]:
    """Helper function for building expected reading or configuration mapping.

    :param val: Value to be wrapped in mapping with "value" as the key.
    :return: The mapping that has wrapped the val with key "value".
    """
    return {"value": val}


def approx_value(value: Any):
    """Allow any value to be compared to another in tests.

    This is needed because numpy arrays give a numpy array back when compared,
    not a bool. This means that you can't ``assert array1==array2``. Numpy
    arrays can be wrapped with `pytest.approx`, but this doesn't work for
    `Table` instances: in this case we use `ApproxTable`.
    """
    return ApproxTable(value) if isinstance(value, Table) else pytest.approx(value)


async def assert_value(signal: SignalR[SignalDatatypeT], value: Any) -> None:
    """Assert that a Signal has the given value.

    :param signal: Signal with get_value.
    :param value: The expected value from the signal.
    """
    actual_value = await signal.get_value()
    assert approx_value(value) == actual_value


async def assert_reading(
    readable: AsyncReadable,
    expected_reading: Mapping[str, Mapping[str, Any]],
    full_match: bool = True,
) -> None:
    """Assert that a readable Device has the given reading.

    :param readable: Device with an async ``read()`` method to get the reading from.
    :param expected_reading: The expected reading from the readable.
    :param full_match: if expected_reading keys set is same as actual keys set.
        true: exact match
        false: expected_reading keys is subset of actual reading keys
    """
    actual_reading = await readable.read()
    _assert_readings_approx_equal(expected_reading, actual_reading, full_match)


def _approx_reading(expected: Mapping[str, Any], actual: Reading) -> Reading:
    ret = dict(
        expected,
        value=approx_value(expected["value"]),
        timestamp=pytest.approx(expected["timestamp"], rel=0.1)
        if "timestamp" in expected
        else actual["timestamp"],
    )
    if "alarm_severity" in actual and "alarm_severity" not in expected:
        ret["alarm_severity"] = actual["alarm_severity"]
    return cast(Reading, ret)


def _assert_readings_approx_equal(
    expected: Mapping[str, Mapping[str, Any]],
    actual: Mapping[str, Reading],
    full_match: bool = True,
):
    # expand the expected keys to include actual if we allow partial matches
    if not full_match:
        expected = dict(actual, **expected)
    # now make them approximate if they are in actual so we get a nicer diff
    approx_expected = {
        k: _approx_reading(v, actual[k]) if k in actual else v
        for k, v in expected.items()
    }
    # now we can compare them
    assert actual == approx_expected


async def assert_configuration(
    configurable: AsyncConfigurable,
    expected_configuration: Mapping[str, Mapping[str, Any]],
    full_match: bool = True,
) -> None:
    """Assert that a configurable Device has the given configuration.

    :param configurable:
        Device with an async ``read_configuration()`` method to get the
        configuration from.
    :param expected_configuration: The expected configuration from the configurable.
    :param full_match: if expected_reading keys set is same as actual keys set.
        true: exact match
        false: expected_reading keys is subset of actual reading keys
    """
    actual_configuration = await configurable.read_configuration()
    _assert_readings_approx_equal(
        expected_configuration, actual_configuration, full_match
    )


async def assert_describe_signal(signal: SignalR, /, **metadata):
    """Assert the describe of a signal matches the expected metadata.

    :param signal: The signal to describe.
    :param metadata: The expected metadata.
    """
    actual_describe = await signal.describe()
    assert list(actual_describe) == [signal.name]
    (actual_datakey,) = actual_describe.values()
    expected_datakey = DataKey(source=signal.source, **metadata)
    assert actual_datakey == expected_datakey


def assert_emitted(docs: Mapping[str, list[dict]], **numbers: int):
    """Assert emitted document generated by running a Bluesky plan.

    :param docs: A mapping of document type -> list of documents that have been emitted.
    :param numbers: The number of each document type expected.

    :example:
    ```python
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))
    RE(my_plan())
    assert_emitted(docs, start=1, descriptor=1, event=1, stop=1)
    ```
    """
    assert list(docs) == list(numbers)
    actual_numbers = {name: len(d) for name, d in docs.items()}
    assert actual_numbers == numbers


def assert_has_calls(device: Device | Signal, calls: Sequence[Any], reset_after=True):
    """Check that the device connected in mock mode has seen the supplied calls.

    :param device: Device or Signal connected in mock mode to check.
    :param calls: Expected sequence of mock calls.
    :param reset_after:
        Whether to reset the mock after assertion so subsequent calls don't see
        the same calls again.

    :example:
    ```python device.trigger() assert_has_calls(
        device, [
            call.device.num_frames.put(1),
            call.device.start_writing.put(None),
        ],
    )
    ```
    """
    mock = get_mock(device)
    assert list(mock.mock_calls) == list(calls)
    if reset_after:
        mock.reset_mock()


class ApproxTable:
    """For approximating two tables are equivalent.

    :param expected: The expected table.
    :param rel: The relative tolerance.
    :param abs: The absolute tolerance.
    :param nan_ok: Whether NaNs are allowed.
    """

    def __init__(self, expected: Table, rel=None, abs=None, nan_ok: bool = False):
        self.expected = expected
        self.rel = rel
        self.abs = abs
        self.nan_ok = nan_ok

    def __eq__(self, value):
        approx_fields = {
            k: pytest.approx(v, self.rel, self.abs, self.nan_ok)
            for k, v in self.expected
        }
        expected = type(self.expected).model_construct(**approx_fields)  # type: ignore
        return expected == value


class MonitorQueue(AbstractContextManager):
    """Monitors a `Signal` and stores its updates."""

    def __init__(self, signal: SignalR):
        self.signal = signal
        self.updates: asyncio.Queue[dict[str, Reading]] = asyncio.Queue()

    async def assert_updates(self, expected_value):
        # Get an update, value and reading
        update = await asyncio.wait_for(self.updates.get(), timeout=5)
        await assert_value(self.signal, expected_value)
        expected_reading = {
            self.signal.name: {
                "value": expected_value,
            }
        }
        await assert_reading(self.signal, expected_reading)
        _assert_readings_approx_equal(expected_reading, update)

    def __enter__(self):
        self.signal.subscribe_reading(self.updates.put_nowait)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.signal.clear_sub(self.updates.put_nowait)


class StatusWatcher(Watcher[T]):
    """Watches an `AsyncStatus`, storing the calls within."""

    def __init__(self, status: WatchableAsyncStatus) -> None:
        self._event = asyncio.Event()
        self.mock = Mock()
        """Mock that stores watcher updates from the status."""
        status.watch(self)

    def __call__(
        self,
        current: T | None = None,
        initial: T | None = None,
        target: T | None = None,
        name: str | None = None,
        unit: str | None = None,
        precision: int | None = None,
        fraction: float | None = None,
        time_elapsed: float | None = None,
        time_remaining: float | None = None,
    ) -> Any:
        self.mock(
            current=current,
            initial=initial,
            target=target,
            name=name,
            unit=unit,
            precision=precision,
            fraction=fraction,
            time_elapsed=time_elapsed,
            time_remaining=time_remaining,
        )
        self._event.set()

    async def wait_for_call(
        self,
        current: T | None = None,
        initial: T | None = None,
        target: T | None = None,
        name: str | None = None,
        unit: str | None = None,
        precision: int | None = None,
        fraction: float | None = None,
        # Any so we can use pytest.approx
        time_elapsed: float | Any = None,
        time_remaining: float | Any = None,
    ):
        await asyncio.wait_for(self._event.wait(), timeout=1)
        assert self.mock.call_count == 1
        assert self.mock.call_args == call(
            current=current,
            initial=initial,
            target=target,
            name=name,
            unit=unit,
            precision=precision,
            fraction=fraction,
            time_elapsed=time_elapsed,
            time_remaining=time_remaining,
        )
        self.mock.reset_mock()
        self._event.clear()



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import sys
from pathlib import Path
from subprocess import check_output

import requests
from sphinx import addnodes, application, environment
from sphinx.ext import intersphinx

import ophyd_async


def missing_reference_handler(
    app: application.Sphinx,
    env: environment.BuildEnvironment,
    node: addnodes.pending_xref,
    contnode,
):
    """Find refs for TypeVars and Unions."""
    target = node["reftarget"]
    if "." in target and node["reftype"] == "class":
        # Try again as `obj` so we pick up Unions, TypeVars and other things
        if target.startswith("ophyd_async"):
            # Pick it up from our domain
            domain = env.domains[node["refdomain"]]
            refdoc = node.get("refdoc")
            return domain.resolve_xref(
                env, refdoc, app.builder, "obj", target, node, contnode
            )
        else:
            # pass it to intersphinx with the right type
            node["reftype"] = "obj"
            return intersphinx.missing_reference(app, env, node, contnode)


def setup(app: application.Sphinx):
    """Add the custom handler to the Sphinx app."""
    app.connect("missing-reference", missing_reference_handler)


# General information about the project.
project = "ophyd-async"
copyright = "2014, Brookhaven National Lab"

# The full version, including alpha/beta/rc tags.
release = ophyd_async.__version__

# The short X.Y version.
if "+" in release:
    # Not on a tag, use branch name
    root = Path(__file__).absolute().parent.parent
    git_branch = check_output("git branch --show-current".split(), cwd=root)
    version = git_branch.decode().strip()
else:
    version = release

extensions = [
    # for diagrams
    "sphinxcontrib.mermaid",
    # Use this for generating API docs
    "autodoc2",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
    # Add a copy button to each code block
    "sphinx_copybutton",
    # For the card element
    "sphinx_design",
    # To make .nojekyll
    "sphinx.ext.githubpages",
    # To make the {ipython} directive
    "IPython.sphinxext.ipython_directive",
    # To syntax highlight "ipython" language code blocks
    "IPython.sphinxext.ipython_console_highlighting",
    # To embed matplotlib plots generated from code
    "matplotlib.sphinxext.plot_directive",
    # To parse markdown
    "myst_parser",
]

# Which package to load and document
autodoc2_packages = [{"path": "../src/ophyd_async", "auto_mode": True}]

# Put them in docs/_api which is git ignored
autodoc2_output_dir = "_api"

# Modules that should have an all...
autodoc2_module_all_regexes = [
    r"ophyd_async\.core",
    r"ophyd_async\.sim",
    r"ophyd_async\.epics\.[^\.]*",
    r"ophyd_async\.tango\.[^\.]*",
    r"ophyd_async\.fastcs\.[^\.]*",
    r"ophyd_async\.plan_stubs",
    r"ophyd_async\.testing",
]

# ... so should have their private modules ignored
autodoc2_skip_module_regexes = [
    x + r"\._.*" for x in autodoc2_module_all_regexes + ["ophyd_async"]
]

# Render with shortened names
autodoc2_render_plugin = "ophyd_async._docs_parser.ShortenedNamesRenderer"

# Don't document private things
autodoc2_hidden_objects = {"private", "dunder", "inherited"}

# We don't have any docstring for __init__, so by separating
# them here we don't get the "Initilize" text that would otherwise be added
autodoc2_class_docstring = "both"

# For some reason annotations are not expanded, this will do here
autodoc2_replace_annotations = [
    ("~PvSuffix.rbv", "ophyd_async.epics.core.PvSuffix.rbv"),
    ("typing_extensions.Self", "typing.Self"),
]

# Which objects to include docstrings for. ‘direct’ means only from objects
# that are not inherited.
autodoc2_docstrings = "all"

# So we can use the ::: syntax and the :param thing: syntax
myst_enable_extensions = ["colon_fence", "fieldlist"]

# If true, Sphinx will warn about all references where the target cannot
# be found.
nitpicky = True

# A list of (type, target) tuples (by default empty) that should be ignored when
# generating warnings in "nitpicky mode". Note that type should include the
# domain name if present. Example entries would be ('py:func', 'int') or
# ('envvar', 'LD_LIBRARY_PATH').
obj_ignore = [
    "ophyd_async.core._derived_signal_backend.RawT",
    "ophyd_async.core._derived_signal_backend.DerivedT",
    "ophyd_async.core._detector.DetectorControllerT",
    "ophyd_async.core._detector.DetectorWriterT",
    "ophyd_async.core._device.DeviceT",
    "ophyd_async.core._device_filler.SignalBackendT",
    "ophyd_async.core._device_filler.DeviceConnectorT",
    "ophyd_async.core._derived_signal_backend.TransformT",
    "ophyd_async.core._mock_signal_backend.MockPutCallback",
    "ophyd_async.core._protocol.C",
    "ophyd_async.core._signal_backend.SignalDatatypeV",
    "ophyd_async.core._status.AsyncStatusBase",
    "ophyd_async.core._utils.P",
    "ophyd_async.core._utils.T",
    "ophyd_async.core._utils.V",
    "ophyd_async.epics.adcore._io.ADBaseIOT",
    "ophyd_async.epics.adcore._io.NDPluginBaseIOT",
    "ophyd_async.tango.core._base_device.T",
    "ophyd_async.tango.core._tango_transport.P",
    "ophyd_async.tango.core._tango_transport.R",
    "ophyd_async.tango.core._tango_transport.TangoProxy",
    "ophyd_async.testing._utils.T",
    "ophyd_async.sim._mirror.TwoJackRaw",
    "ophyd_async.sim._mirror.TwoJackDerived",
    "non_zero",
    "0.1",
    "1.0",
    "bluesky.protocols.T_co",
]
nitpick_ignore = []
for var in obj_ignore:
    nitpick_ignore.append(("py:class", var))
    nitpick_ignore.append(("py:obj", var))
# Ignore classes in modules with no intersphinx
nitpick_ignore_regex = [
    (r"py:.*", r"pydantic\..*"),
    (r"py:.*", r"tango\..*"),
]

# Output graphviz directive produced images in a scalable format
graphviz_output_format = "svg"

# The name of a reST role (builtin or Sphinx extension) to use as the default
# role, that is, for text marked up `like this`
default_role = "any"

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# These patterns also affect html_static_path and html_extra_path
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Example configuration for intersphinx: refer to the Python standard library.
# This means you can link things like `str` and `asyncio` to the relevant
# docs in the python documentation.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "bluesky": ("https://blueskyproject.io/bluesky/main", None),
    "scanspec": ("https://blueskyproject.io/scanspec/main", None),
    "numpy": ("https://numpy.org/devdocs/", None),
    "event-model": ("https://blueskyproject.io/event-model/main", None),
    "pytest": ("https://docs.pytest.org/en/stable/", None),
}

# A dictionary of graphviz graph attributes for inheritance diagrams.
inheritance_graph_attrs = {"rankdir": "TB"}

# Ignore localhost links for periodic check that links in docs are valid
linkcheck_ignore = [r"http://localhost:\d+/"]

# Set copy-button to ignore python and bash prompts
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
github_repo = "ophyd-async"
github_user = "bluesky"
switcher_json = "https://blueskyproject.io/ophyd-async/switcher.json"
switcher_exists = requests.get(switcher_json).ok
if not switcher_exists:
    print(
        "*** Can't read version switcher, is GitHub pages enabled? \n"
        "    Once Docs CI job has successfully run once, set the "
        "Github pages source branch to be 'gh-pages' at:\n"
        f"    https://github.com/{github_user}/{github_repo}/settings/pages",
        file=sys.stderr,
    )

# Theme options for pydata_sphinx_theme
# We don't check switcher because there are 3 possible states for a repo:
# 1. New project, docs are not published so there is no switcher
# 2. Existing project with latest skeleton, switcher exists and works
# 3. Existing project with old skeleton that makes broken switcher,
#    switcher exists but is broken
# Point 3 makes checking switcher difficult, because the updated skeleton
# will fix the switcher at the end of the docs workflow, but never gets a chance
# to complete as the docs build warns and fails.
html_theme_options = {
    "use_edit_page_button": True,
    "github_url": f"https://github.com/{github_user}/{github_repo}",
    "icon_links": [
        {
            "name": "PyPI",
            "url": f"https://pypi.org/project/{project}",
            "icon": "fas fa-cube",
        },
    ],
    "switcher": {
        "json_url": switcher_json,
        "version_match": version,
    },
    "check_switcher": False,
    "navbar_end": ["theme-switcher", "icon-links", "version-switcher"],
    "external_links": [
        {
            "name": "Bluesky Project",
            "url": "https://blueskyproject.io",
        },
    ],
    "navigation_with_keys": False,
    "show_toc_level": 3,
}

# A dictionary of values to pass into the template engine’s context for all pages
html_context = {
    "github_user": github_user,
    "github_repo": project,
    "github_version": version,
    "doc_path": "docs",
}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Logo
html_logo = "images/ophyd-async-logo.svg"
html_favicon = "images/ophyd-favicon.svg"

# Custom CSS
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Where to put Ipython savefigs
ipython_savefig_dir = "../build/savefig"



from ophyd_async.core import DEFAULT_TIMEOUT, DetectorArmLogic, wait_for_value

from ._io import EigerDetectorIO


class EigerArmLogic(DetectorArmLogic):
    def __init__(self, detector: EigerDetectorIO) -> None:
        self.detector = detector

    async def arm(self):
        await self.detector.arm.trigger()

    async def wait_for_idle(self):
        await wait_for_value(self.detector.state, "idle", timeout=DEFAULT_TIMEOUT)

    async def disarm(self):
        await self.detector.disarm.trigger()



"""Used for tutorial `Using Devices`."""

# Import bluesky and ophyd
from pathlib import PurePath
from tempfile import mkdtemp

import bluesky.plan_stubs as bps  # noqa: F401
import bluesky.plans as bp  # noqa: F401
import bluesky.preprocessors as bpp  # noqa: F401
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.run_engine import RunEngine, autoawait_in_bluesky_event_loop

from ophyd_async import sim
from ophyd_async.core import StaticPathProvider, UUIDFilenameProvider, init_devices

# Create a run engine and make ipython use it for `await` commands
RE = RunEngine(call_returns_result=True)
autoawait_in_bluesky_event_loop()

# Add a callback for plotting
bec = BestEffortCallback()
RE.subscribe(bec)

# Make a pattern generator that uses the motor positions
# to make a test pattern. This simulates the real life process
# of X-ray scattering off a sample
pattern_generator = sim.PatternGenerator()

# Make a path provider that makes UUID filenames within a static
# temporary directory
path_provider = StaticPathProvider(UUIDFilenameProvider(), PurePath(mkdtemp()))

# All Devices created within this block will be
# connected and named at the end of the with block
with init_devices():
    # Create a sample stage with X and Y motors that report their positions
    # to the pattern generator
    stage = sim.SimStage(pattern_generator)
    # Make a detector device that gives the point value of the pattern generator
    # when triggered
    pdet = sim.SimPointDetector(pattern_generator)
    # Make a detector device that gives a gaussian blob with intensity based
    # on the point value of the pattern generator when triggered
    bdet = sim.SimBlobDetector(path_provider, pattern_generator)



from ophyd_async.core import (
    Device,
    SignalR,
    SignalRW,
    SignalX,
    StrictEnum,
)


class EigerTriggerMode(StrictEnum):
    INTERNAL = "ints"
    EDGE = "exts"
    GATE = "exte"


class EigerMonitorIO(Device):
    """Driver for Eiger Monitor subsystem.

    This mirrors the interface provided by https://media.dectris.com/SIMPLON_APIReference_v1p6.pdf#page=25
    """

    pass


class EigerStreamIO(Device):
    """Driver for Eiger Stream subsystem.

    This mirrors the interface provided by https://media.dectris.com/SIMPLON_APIReference_v1p6.pdf#page=32
    """

    pass


class EigerDetectorIO(Device):
    """Driver for Eiger Detector subsystem.

    This mirrors the interface provided by https://media.dectris.com/SIMPLON_APIReference_v1p6.pdf#page=17
    """

    bit_depth_image: SignalR[int]
    state: SignalR[str]
    count_time: SignalRW[float]
    frame_time: SignalRW[float]
    nimages: SignalRW[int]
    ntrigger: SignalRW[int]
    nexpi: SignalRW[int] | None
    trigger_mode: SignalRW[str]
    roi_mode: SignalRW[str]
    photon_energy: SignalRW[float]
    beam_center_x: SignalRW[float]
    beam_center_y: SignalRW[float]
    detector_distance: SignalRW[float]
    omega_start: SignalRW[float]
    omega_increment: SignalRW[float]
    arm: SignalX
    disarm: SignalX
    trigger: SignalX



from unittest.mock import ANY

import pytest

from ophyd_async.core import (
    DetectorDataLogic,
    StaticPathProvider,
    TriggerInfo,
    callback_on_mock_put,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adaravis, adcore, adsimdetector
from ophyd_async.epics.core import epics_signal_r
from ophyd_async.testing import assert_configuration, assert_reading


async def test_step_scan_hdf_detector_with_stats_and_temp(
    static_path_provider: StaticPathProvider,
):
    stat = adcore.NDStatsIO("PREFIX:STATS:")
    async with init_devices(mock=True):
        det = adsimdetector.SimDetector(
            "PREFIX:", static_path_provider, plugins={"stats": stat}
        )
        temp = epics_signal_r(float, "SAMP:TEMP:RBV")
    ndattributes = [
        adcore.NDAttributeParam(
            name="det-sum",
            param="TOTAL",
            datatype=adcore.NDAttributeDataType.DOUBLE,
            description="Sum of the array",
        ),
        adcore.NDAttributePv(
            name="sample-temp",
            signal=temp,
            dbrtype=adcore.NDAttributePvDbrType.DBR_DOUBLE,
            description="Temperature of the sample",
        ),
    ]
    set_mock_value(stat.nd_attributes_file, adcore.ndattributes_to_xml(ndattributes))
    writer = det.get_plugin("writer", adcore.NDFileHDF5IO)
    set_mock_value(det.driver.acquire_period, 0.1)
    set_mock_value(det.driver.acquire_time, 0.05)
    set_mock_value(det.driver.array_size_x, 1024)
    set_mock_value(det.driver.array_size_y, 768)

    await det.stage()
    assert await det.driver.wait_for_plugins.get_value() is False
    config_description = await det.describe_configuration()
    assert config_description == {
        "det-driver-acquire_period": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://PREFIX:cam1:AcquirePeriod_RBV",
        },
        "det-driver-acquire_time": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://PREFIX:cam1:AcquireTime_RBV",
        },
    }
    await assert_configuration(
        det,
        {
            "det-driver-acquire_period": {"value": 0.1},
            "det-driver-acquire_time": {"value": 0.05},
        },
    )
    # Check we need to call prepare or trigger before we can describe
    with pytest.raises(RuntimeError, match="Prepare not run"):
        await det.describe()
    # Need to tell it the path is valid
    set_mock_value(writer.file_path_exists, True)
    # When arm is pressed, then make a single frame
    callback_on_mock_put(
        det.driver.acquire, lambda v: set_mock_value(writer.num_captured, 1)
    )
    # Trigger a single frame then describe, get hints and read
    await det.trigger()
    path_info = static_path_provider()
    description = await det.describe()
    uri = f"file://localhost/{path_info.directory_path.as_posix().lstrip('/')}/{path_info.filename}.h5"
    assert description == {
        "det": {
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
            "shape": [1, 768, 1024],
            "source": uri,
        },
        "det-sum": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
            "shape": [1],
            "source": uri,
        },
        "sample-temp": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
            "shape": [1],
            "source": "ca://SAMP:TEMP:RBV",
        },
    }
    assert det.hints == {"fields": ["det"]}
    await assert_reading(det, {})
    docs = [doc async for doc in det.collect_asset_docs()]
    sr = docs[:3]
    sd = docs[3:]
    assert sr == [
        (
            "stream_resource",
            {
                "data_key": "det",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (1, 768, 1024),
                    "dataset": "/entry/data/data",
                },
                "uid": ANY,
                "uri": uri,
            },
        ),
        (
            "stream_resource",
            {
                "data_key": "det-sum",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (16384,),
                    "dataset": "/entry/instrument/NDAttributes/det-sum",
                },
                "uid": ANY,
                "uri": uri,
            },
        ),
        (
            "stream_resource",
            {
                "data_key": "sample-temp",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (16384,),
                    "dataset": "/entry/instrument/NDAttributes/sample-temp",
                },
                "uid": ANY,
                "uri": uri,
            },
        ),
    ]
    assert sd == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 1},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[i][1]["uid"],
                "uid": ANY,
            },
        )
        for i in range(3)
    ]
    # Check we can prepare and do a second one
    await det.prepare(TriggerInfo(exposure_timeout=0.1))
    callback_on_mock_put(
        det.driver.acquire, lambda v: set_mock_value(writer.num_captured, 2)
    )
    await det.trigger()
    readings = await det.read()
    assert readings == {}
    docs = [doc async for doc in det.collect_asset_docs()]
    assert docs == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 1, "stop": 2},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[i][1]["uid"],
                "uid": ANY,
            },
        )
        for i in range(3)
    ]
    # And if we trigger again without updating num_captured then we timeout
    with pytest.raises(
        TimeoutError, match="Timeout Error while waiting 0.1s to update"
    ):
        await det.trigger()


async def test_step_scan_tiff_detector(
    static_path_provider: StaticPathProvider,
):
    async with init_devices(mock=True):
        det = adsimdetector.SimDetector(
            "PREFIX:", static_path_provider, writer_type=adcore.ADWriterType.TIFF
        )

    writer = det.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(det.driver.array_size_x, 1024)
    set_mock_value(det.driver.array_size_y, 768)

    await det.stage()
    # Need to tell it the path is valid
    set_mock_value(writer.file_path_exists, True)
    # When arm is pressed, then make a single frame
    callback_on_mock_put(
        det.driver.acquire, lambda v: set_mock_value(writer.num_captured, 1)
    )
    # Trigger a single frame then describe and read
    await det.trigger()
    path_info = static_path_provider()
    description = await det.describe()
    uri_dir = f"file://localhost/{path_info.directory_path.as_posix().lstrip('/')}/"
    assert description == {
        "det": {
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
            "shape": [1, 768, 1024],
            "source": uri_dir,
        },
    }
    assert det.hints == {"fields": ["det"]}
    readings = await det.read()
    assert readings == {}
    docs = [doc async for doc in det.collect_asset_docs()]
    sr = docs[:1]
    sd = docs[1:]
    assert sr == [
        (
            "stream_resource",
            {
                "data_key": "det",
                "mimetype": "multipart/related;type=image/tiff",
                "parameters": {
                    "chunk_shape": (1, 768, 1024),
                    "template": "ophyd_async_tests_{:06d}.tiff",
                },
                "uid": ANY,
                "uri": uri_dir,
            },
        ),
    ]
    assert sd == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 1},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[0][1]["uid"],
                "uid": ANY,
            },
        ),
    ]


async def test_flyscan_aravis_detector(static_path_provider: StaticPathProvider):
    async with init_devices(mock=True):
        det = adaravis.AravisDetector("PREFIX:", static_path_provider)

    writer = det.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(det.driver.model, "A funny model")
    set_mock_value(det.driver.acquire_period, 0.1)
    set_mock_value(det.driver.acquire_time, 0.05)
    set_mock_value(det.driver.array_size_x, 1024)
    set_mock_value(det.driver.array_size_y, 768)
    await det.stage()
    # Check the model is in the configuration too
    config_description = await det.describe_configuration()
    assert config_description == {
        "det-driver-acquire_period": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://PREFIX:cam1:AcquirePeriod_RBV",
        },
        "det-driver-acquire_time": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://PREFIX:cam1:AcquireTime_RBV",
        },
        "det-driver-model": {
            "dtype": "string",
            "dtype_numpy": "|S40",
            "shape": [],
            "source": "mock+ca://PREFIX:cam1:Model_RBV",
        },
    }
    await assert_configuration(
        det,
        {
            "det-driver-acquire_period": {"value": 0.1},
            "det-driver-acquire_time": {"value": 0.05},
            "det-driver-model": {"value": "A funny model"},
        },
    )
    # Prepare for N frames
    set_mock_value(writer.file_path_exists, True)
    await det.prepare(TriggerInfo(collections_per_event=3, number_of_events=5))
    # Check it is making 3 collections in each event
    path_info = static_path_provider()
    description = await det.describe()
    uri = f"file://localhost/{path_info.directory_path.as_posix().lstrip('/')}/{path_info.filename}.h5"
    assert description == {
        "det": {
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
            "shape": [3, 768, 1024],
            "source": uri,
        },
    }
    assert det.hints == {"fields": ["det"]}
    # Kick it off and start it completing
    await det.kickoff()
    status = det.complete()
    # Check it makes nothing
    docs = [doc async for doc in det.collect_asset_docs()]
    assert docs == []
    # Make 2 and 1/3 events and check they are produced
    set_mock_value(writer.num_captured, 7)
    docs = [doc async for doc in det.collect_asset_docs()]
    assert docs == [
        (
            "stream_resource",
            {
                "data_key": "det",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (1, 768, 1024),
                    "dataset": "/entry/data/data",
                },
                "uid": ANY,
                "uri": uri,
            },
        ),
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 2},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": ANY,
                "uid": ANY,
            },
        ),
    ]
    assert not status.done
    # Update and check it completes
    set_mock_value(writer.num_captured, 15)
    docs = [doc async for doc in det.collect_asset_docs()]
    assert docs == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 2, "stop": 5},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": ANY,
                "uid": ANY,
            },
        ),
    ]
    assert status.done
    # Check that another kickoff without prepare is not allowed
    with pytest.raises(
        RuntimeError,
        match="Kickoff requested 15:30, but detector was only prepared up to 15",
    ):
        await det.kickoff()
    # But preparing again is ok
    await det.prepare(TriggerInfo(collections_per_event=3, number_of_events=3))
    await det.kickoff()
    # Completing adds indexes to the same resource
    status = det.complete()
    set_mock_value(writer.num_captured, 24)
    docs = [doc async for doc in det.collect_asset_docs()]
    assert docs == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 5, "stop": 8},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": ANY,
                "uid": ANY,
            },
        ),
    ]


async def test_2_rois_with_hdf(tmp_path):
    path_provider = StaticPathProvider(
        lambda device_name=None: str(device_name), tmp_path
    )
    driver = adcore.ADBaseIO("PREFIX:DRV:")
    rois: list[adcore.NDROIIO] = []
    hdfs: list[adcore.NDFileHDF5IO] = []
    logics: list[DetectorDataLogic] = []
    for i in range(1, 3):
        roi = adcore.NDROIIO(f"PREFIX:ROI{i}")
        hdf = adcore.NDFileHDF5IO(f"PREFIX:HDF{i}")
        rois.append(roi)
        hdfs.append(hdf)
        logics.append(
            adcore.ADHDFDataLogic(
                description=adcore.NDArrayDescription(
                    shape_signals=(roi.size_y, roi.size_x),
                    data_type_signal=driver.data_type,
                    color_mode_signal=driver.color_mode,
                ),
                path_provider=path_provider,
                driver=driver,
                writer=hdf,
                datakey_suffix=f"-roi{i}",
            )
        )
    async with init_devices(mock=True):
        det = adcore.AreaDetector(
            driver,
            arm_logic=adcore.ADArmLogic(driver),
            writer_type=None,
            plugins={
                "hdf1": hdfs[0],
                "hdf2": hdfs[1],
                "roi1": rois[0],
                "roi2": rois[1],
            },
        )
        det.add_detector_logics(*logics)
    await det.stage()

    # When arm is pressed, then make a single frame on each HDF
    def publish_captured(v):
        for hdf in hdfs:
            set_mock_value(hdf.num_captured, 1)

    callback_on_mock_put(det.driver.acquire, publish_captured)
    # Setup the size of the rois and say the directory exists
    set_mock_value(rois[0].size_x, 400)
    set_mock_value(rois[0].size_y, 300)
    set_mock_value(rois[1].size_x, 200)
    set_mock_value(rois[1].size_y, 100)
    set_mock_value(hdfs[0].file_path_exists, True)
    set_mock_value(hdfs[1].file_path_exists, True)
    # Trigger a single frame then describe and read
    await det.trigger()
    description = await det.describe()
    path_info = path_provider()
    uri1 = f"file://localhost/{path_info.directory_path.as_posix().lstrip('/')}/det-roi1.h5"
    uri2 = f"file://localhost/{path_info.directory_path.as_posix().lstrip('/')}/det-roi2.h5"
    assert description == {
        "det-roi1": {
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
            "shape": [1, 300, 400],
            "source": uri1,
        },
        "det-roi2": {
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
            "shape": [1, 100, 200],
            "source": uri2,
        },
    }
    assert det.hints == {"fields": ["det-roi1", "det-roi2"]}
    docs = [doc async for doc in det.collect_asset_docs()]
    sr = docs[0:4:2]
    sd = docs[1:5:2]
    assert sr == [
        (
            "stream_resource",
            {
                "data_key": "det-roi1",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (1, 300, 400),
                    "dataset": "/entry/data/data",
                },
                "uid": ANY,
                "uri": uri1,
            },
        ),
        (
            "stream_resource",
            {
                "data_key": "det-roi2",
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (1, 100, 200),
                    "dataset": "/entry/data/data",
                },
                "uid": ANY,
                "uri": uri2,
            },
        ),
    ]
    assert sd == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 1},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[0][1]["uid"],
                "uid": ANY,
            },
        ),
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 1},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[1][1]["uid"],
                "uid": ANY,
            },
        ),
    ]


async def test_simdetector_with_stats_signal():
    stat = adcore.NDStatsIO("PREFIX:STAT:")
    async with init_devices(mock=True):
        det = adsimdetector.SimDetector(
            "PREFIX:", writer_type=None, plugins={"stats": stat}
        )
        det.add_detector_logics(adcore.PluginSignalDataLogic(det.driver, stat.total))
    set_mock_value(stat.total, 1.8)
    await det.stage()
    assert await det.driver.wait_for_plugins.get_value() is False
    await det.trigger()
    assert await det.driver.wait_for_plugins.get_value() is True
    description = await det.describe()
    assert description == {
        "det-stats-total": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://PREFIX:STAT:Total_RBV",
        }
    }
    await assert_reading(det, {"det-stats-total": {"value": 1.8}})
    assert det.hints == {"fields": ["det-stats-total"]}



from ophyd_async.core import Device

from ._epics_connector import EpicsDeviceConnector
from ._pvi_connector import PviDeviceConnector


class EpicsDevice(Device):
    """Baseclass to allow child signals to be created declaratively."""

    def __init__(self, prefix: str, with_pvi: bool = False, name: str = ""):
        if with_pvi:
            connector = PviDeviceConnector(prefix)
        else:
            connector = EpicsDeviceConnector(prefix)
        super().__init__(name=name, connector=connector)



import asyncio
import logging
import re
from collections.abc import Sequence
from typing import Any

import numpy as np
import numpy.typing as npt
import pytest
from tango import AttrDataFormat, CmdArgType, DevState
from tango.asyncio import DeviceProxy
from tango.asyncio_executor import (
    AsyncioExecutor,
    get_global_executor,
)
from test_base_device import TestDevice
from test_tango_signals import make_backend

from ophyd_async.core import (
    Array1D,
    NotConnectedError,
    StrictEnum,
)
from ophyd_async.tango.core import (
    AttributeProxy,
    CommandProxy,
    TangoDoubleStringTable,
    TangoLongStringTable,
    TangoSignalBackend,
    ensure_proper_executor,
    get_dtype_extended,
    get_full_attr_trl,
    get_python_type,
    get_tango_trl,
    try_to_cast_as_float,
)
from ophyd_async.tango.testing import TestConfig


# --------------------------------------------------------------------
async def prepare_device(trl: str, pv: str, put_value: Any) -> None:
    proxy = await DeviceProxy(trl)
    setattr(proxy, pv, put_value)


# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def tango_test_device(subprocess_helper):
    with subprocess_helper(
        [{"class": TestDevice, "devices": [{"name": "test/device/1"}]}]
    ) as context:
        yield context.trls["test/device/1"]


# --------------------------------------------------------------------
class HelperClass:
    @ensure_proper_executor
    async def mock_func(self):
        return "executed"


# Test function
@pytest.mark.asyncio
async def test_ensure_proper_executor():
    # Instantiate the helper class and call the decorated method
    helper_instance = HelperClass()
    result = await helper_instance.mock_func()

    # Assertions
    assert result == "executed"
    assert isinstance(get_global_executor(), AsyncioExecutor)


# --------------------------------------------------------------------


@pytest.mark.parametrize(
    "tango_type, tango_format, expected",
    [
        # Scalar types
        (CmdArgType.DevVoid, AttrDataFormat.SCALAR, None),
        (CmdArgType.DevBoolean, AttrDataFormat.SCALAR, bool),
        (CmdArgType.DevShort, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevLong, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevLong64, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevFloat, AttrDataFormat.SCALAR, float),
        (CmdArgType.DevDouble, AttrDataFormat.SCALAR, float),
        (CmdArgType.DevUShort, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevULong, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevULong64, AttrDataFormat.SCALAR, int),
        (CmdArgType.DevString, AttrDataFormat.SCALAR, str),
        (CmdArgType.DevEncoded, AttrDataFormat.SCALAR, str),
        (CmdArgType.DevEnum, AttrDataFormat.SCALAR, StrictEnum),
        (CmdArgType.DevState, AttrDataFormat.SCALAR, StrictEnum),
        (CmdArgType.ConstDevString, AttrDataFormat.SCALAR, str),
        (CmdArgType.DevVarBooleanArray, AttrDataFormat.SCALAR, bool),
        (CmdArgType.DevUChar, AttrDataFormat.SCALAR, int),
        # Array types
        (CmdArgType.DevVarCharArray, AttrDataFormat.SPECTRUM, Sequence[str]),
        (CmdArgType.DevVarShortArray, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarShortArray, AttrDataFormat.IMAGE, npt.NDArray[int]),
        (CmdArgType.DevVarLongArray, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarLongArray, AttrDataFormat.IMAGE, npt.NDArray[int]),
        (CmdArgType.DevVarFloatArray, AttrDataFormat.SPECTRUM, Array1D[float]),
        (CmdArgType.DevVarFloatArray, AttrDataFormat.IMAGE, npt.NDArray[float]),
        (CmdArgType.DevVarDoubleArray, AttrDataFormat.SPECTRUM, Array1D[float]),
        (CmdArgType.DevVarDoubleArray, AttrDataFormat.IMAGE, npt.NDArray[float]),
        (CmdArgType.DevVarUShortArray, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarUShortArray, AttrDataFormat.IMAGE, npt.NDArray[int]),
        (CmdArgType.DevVarULongArray, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarULongArray, AttrDataFormat.IMAGE, npt.NDArray[int]),
        (CmdArgType.DevVarLong64Array, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarLong64Array, AttrDataFormat.IMAGE, npt.NDArray[int]),
        (CmdArgType.DevVarULong64Array, AttrDataFormat.SPECTRUM, Array1D[int]),
        (CmdArgType.DevVarULong64Array, AttrDataFormat.IMAGE, npt.NDArray[int]),
        # String array types
        (CmdArgType.DevVarStringArray, AttrDataFormat.SPECTRUM, Sequence[str]),
        (CmdArgType.DevVarStringArray, AttrDataFormat.IMAGE, Sequence[Sequence[str]]),
        (
            CmdArgType.DevVarLongStringArray,
            AttrDataFormat.SPECTRUM,
            TangoLongStringTable,
        ),
        (
            CmdArgType.DevVarDoubleStringArray,
            AttrDataFormat.SPECTRUM,
            TangoDoubleStringTable,
        ),
        # Bad type
        (float, AttrDataFormat.SCALAR, (False, float, "number")),
        # Bad format
        (float, "bad_format", (False, float, "format")),
    ],
)
def test_get_python_type(tango_type, tango_format, expected):
    config = TestConfig()
    config.data_format = tango_format
    config.data_type = tango_type
    if tango_type is CmdArgType.DevEnum:
        config.enum_labels = ["A", "B", "C"]
        py_type = get_python_type(config)
        assert issubclass(py_type, StrictEnum)
        assert [e.name for e in py_type] == ["A", "B", "C"]
    elif tango_type is CmdArgType.DevState:
        py_type = get_python_type(config)
        assert issubclass(py_type, StrictEnum)
        assert [e.name for e in py_type] == list(DevState.names.keys())
    elif tango_type is not float:
        if (
            tango_format is AttrDataFormat.IMAGE
            and tango_type is CmdArgType.DevVarStringArray
        ):
            with pytest.raises(TypeError) as exc:
                get_python_type(config)
            assert str(exc.value) == "Images of type str or enum are not supported"
        else:
            assert get_python_type(config) == expected
        return
    else:
        if tango_format == "bad_format":
            with pytest.raises(TypeError) as exc:
                get_python_type(config)
            assert str(exc.value) == "Unknown TangoFormat"
            return
        # get_python_type should raise a TypeError
        with pytest.raises(TypeError) as exc:
            get_python_type(config)
        assert str(exc.value) == "Unknown TangoType: <class 'float'>"


# --------------------------------------------------------------------
def test_try_to_cast_as_float():
    # Test with a valid float value
    result = try_to_cast_as_float(3.14)
    assert result == 3.14

    # Test with a valid integer value
    result = try_to_cast_as_float(42)
    assert result == 42.0

    # Test with a string that can be converted to float
    result = try_to_cast_as_float("2.718")
    assert result == 2.718

    # Test with a string that cannot be converted to float
    result = try_to_cast_as_float("not_a_number")
    assert result is None


# --------------------------------------------------------------------
@pytest.mark.parametrize(
    "datatype, expected",
    [
        (npt.NDArray[np.float64], np.dtype("float64")),
        (npt.NDArray[np.int8], np.dtype("int8")),
        (npt.NDArray[np.uint8], np.dtype("uint8")),
        (npt.NDArray[np.int32], np.dtype("int32")),
        (npt.NDArray[np.int64], np.dtype("int64")),
        (npt.NDArray[np.uint16], np.dtype("uint16")),
        (npt.NDArray[np.uint32], np.dtype("uint32")),
        (npt.NDArray[np.uint64], np.dtype("uint64")),
        (npt.NDArray[np.bool_], np.dtype("bool")),
        (npt.NDArray[DevState], CmdArgType.DevState),
        (npt.NDArray[np.str_], np.dtype("str")),
        (npt.NDArray[np.float32], np.dtype("float32")),
        (npt.NDArray[np.complex64], np.dtype("complex64")),
        (npt.NDArray[np.complex128], np.dtype("complex128")),
    ],
)
def test_get_dtype_extended(datatype, expected):
    assert get_dtype_extended(datatype) == expected


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "attr_name, proxy_needed, expected_type, should_raise",
    [
        ("justvalue", True, AttributeProxy, False),
        ("justvalue", False, AttributeProxy, False),
        ("clear", True, CommandProxy, False),
        ("clear", False, CommandProxy, False),
        ("nonexistent", True, None, True),
    ],
)
@pytest.mark.timeout(10)
async def test_get_tango_trl(
    tango_test_device, attr_name, proxy_needed, expected_type, should_raise
):
    trl = get_full_attr_trl(tango_test_device, attr_name)
    assert re.match(
        r"tango://[a-zA-Z0-9\.-_:]*/test/device/1/" + attr_name + r"#dbase=no", trl
    )
    # tango_test_device is of form tango://127.0.0.1:<port>/test/device/1#dbase=no
    proxy = await DeviceProxy(tango_test_device) if proxy_needed else None
    if should_raise:
        with pytest.raises(RuntimeError):
            await get_tango_trl(trl, proxy, 1)
    else:
        await asyncio.sleep(0.1)
        result = await get_tango_trl(trl, proxy, 1)
        assert isinstance(result, expected_type)


# --------------------------------------------------------------------


@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["justvalue", "array"])
async def test_attribute_proxy_get(tango_test_device, attr):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    val = None
    val = await attr_proxy.get()
    assert val is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "attr",
    ["justvalue", "array"],
)
async def test_attribute_proxy_put(tango_test_device, attr):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    old_value = await attr_proxy.get()
    new_value = old_value + 1
    await attr_proxy.put(new_value)
    await asyncio.sleep(
        0.1
    )  # for some reason this is required otherwise justvalue fails???
    updated_value = await attr_proxy.get()
    if isinstance(new_value, np.ndarray):
        assert np.all(updated_value == new_value)
    else:
        assert updated_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_proxy_put_force_timeout(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "slow_attribute")
    with pytest.raises(TimeoutError) as exc:
        await attr_proxy.put(3.0, timeout=0.1)
    assert "Timeout" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_proxy_put_exceptions(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "raise_exception_attr")
    with pytest.raises(RuntimeError) as exc:
        await attr_proxy.put(3.0)
    assert "device failure" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "attr, new_value", [("justvalue", 10), ("array", np.array([[2, 3, 4], [5, 6, 7]]))]
)
@pytest.mark.timeout(4.7)
async def test_attribute_proxy_get_w_value(tango_test_device, attr, new_value):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    await attr_proxy.put(new_value)
    await asyncio.sleep(1.0)
    attr_proxy_value = await attr_proxy.get()
    if isinstance(new_value, np.ndarray):
        assert np.all(attr_proxy_value == new_value)
    else:
        assert attr_proxy_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_get_config(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    config = await attr_proxy.get_config()
    assert config.writable is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_get_reading(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    reading = await attr_proxy.get_reading()
    assert reading["value"] is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_has_subscription(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    expected = bool(attr_proxy._callback)
    has_subscription = attr_proxy.has_subscription()
    assert has_subscription is expected


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_set_polling(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    attr_proxy.set_polling(True, 0.1, 1, 0.1)
    assert attr_proxy._allow_polling
    assert attr_proxy._polling_period == 0.1
    assert attr_proxy._abs_change == 1
    assert attr_proxy._rel_change == 0.1
    attr_proxy.set_polling(False)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(3)
async def test_attribute_poll(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "floatvalue")
    attr_proxy.support_events = False

    def callback(reading):
        nonlocal val
        val = reading["value"]

    def bad_callback():
        pass

    # Test polling with absolute change
    val = None

    attr_proxy.set_polling(True, 0.1, 1, 1.0)
    attr_proxy.subscribe_callback(callback)
    current_value = await attr_proxy.get()
    new_value = current_value + 2
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period)
    assert val is not None
    attr_proxy.unsubscribe_callback()

    # Test polling with relative change
    val = None
    attr_proxy.set_polling(True, 0.1, 100, 0.1)
    attr_proxy.subscribe_callback(callback)
    current_value = await attr_proxy.get()
    new_value = current_value * 2
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period)
    assert val is not None
    attr_proxy.unsubscribe_callback()

    # Test polling with small changes. This should not update last_reading
    attr_proxy.set_polling(True, 0.1, 100, 1.0)
    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    current_value = await attr_proxy.get()
    new_value = current_value + 1
    val = None
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period * 2)
    assert val is None
    attr_proxy.unsubscribe_callback()

    # Test polling with bad callback
    attr_proxy.subscribe_callback(bad_callback)
    await asyncio.sleep(0.2)
    assert "Could not poll the attribute" in str(attr_proxy.exception)
    attr_proxy.unsubscribe_callback()


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["array", "label"])
@pytest.mark.timeout(4.2)
async def test_attribute_poll_stringsandarrays(tango_test_device, attr):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    attr_proxy.support_events = False

    def callback(reading):
        nonlocal val
        val = reading["value"]

    val = None
    attr_proxy.set_polling(True, 0.1)
    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    assert val is not None
    if isinstance(val, np.ndarray):
        await attr_proxy.put(np.array([[2, 3, 4], [5, 6, 7]]))
        await asyncio.sleep(0.5)
        assert np.all(val == np.array([[2, 3, 4], [5, 6, 7]]))
    if isinstance(val, str):
        await attr_proxy.put("new label")
        await asyncio.sleep(0.5)
        assert val == "new label"

    assert attr_proxy._poll_task
    attr_proxy.unsubscribe_callback()


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_poll_exceptions(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    # Try to poll a non-existent attribute
    attr_proxy = AttributeProxy(device_proxy, "nonexistent")
    attr_proxy.support_events = False
    attr_proxy.set_polling(True, 0.1)

    def callback(reading, value):
        pass

    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    assert "Could not poll the attribute" in str(attr_proxy.exception)
    attr_proxy.unsubscribe_callback()


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_proxy_put_wait(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "echo")

    cmd_proxy._last_reading = None
    await cmd_proxy.put("test_message")
    assert cmd_proxy._last_reading["value"] == "test_message"

    # Force timeout
    cmd_proxy = CommandProxy(device_proxy, "slow_command")
    cmd_proxy._last_reading = None
    with pytest.raises(TimeoutError) as exc:
        await cmd_proxy.put(None, timeout=0.1)
    assert "command failed" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(3.2)
async def test_command_proxy_put_nowait(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "slow_command")

    # Reply before timeout
    cmd_proxy._last_reading = None
    status = cmd_proxy.put(None, timeout=0.5)
    assert cmd_proxy._last_reading is None
    await status
    assert cmd_proxy._last_reading["value"] == "Completed slow command"

    # Timeout
    cmd_proxy._last_reading = None
    status = cmd_proxy.put(None, timeout=0.1)
    with pytest.raises(TimeoutError) as exc:
        await status
    assert "Timeout" in str(exc.value)

    # No timeout
    cmd_proxy._last_reading = None
    status = cmd_proxy.put(None)
    assert cmd_proxy._last_reading is None
    await status
    assert cmd_proxy._last_reading["value"] == "Completed slow command"


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("wait", [True, False])
async def test_command_proxy_put_exceptions(tango_test_device, wait):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "raise_exception_cmd")
    await cmd_proxy.connect()
    with pytest.raises(RuntimeError) as exc:
        await cmd_proxy.put(None)
    assert "device failure" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "echo")
    await cmd_proxy.connect()
    await cmd_proxy.put("test_message", timeout=1.0)
    value = await cmd_proxy.get()
    assert value == "test_message"


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get_config(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    config = await cmd_proxy.get_config()
    assert config.out_type is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get_reading(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "echo")
    await cmd_proxy.connect()
    await cmd_proxy.put("test_message", timeout=1.0)
    reading = await cmd_proxy.get_reading()
    assert reading["value"] == "test_message"


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_set_polling(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    cmd_proxy.set_polling(True, 0.1)
    # Set polling in the command proxy currently does nothing
    assert True


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_init(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "justvalue")
    transport = await make_backend(float, source, connect=False)
    assert transport is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_source(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "justvalue")
    transport = await make_backend(int, source)
    transport_source = transport.source("", True)
    assert transport_source == source


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_datatype_allowed(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    backend = await make_backend(float, source)

    assert backend.datatype_allowed(int)
    assert backend.datatype_allowed(float)
    assert backend.datatype_allowed(str)
    assert backend.datatype_allowed(bool)
    assert backend.datatype_allowed(np.ndarray)
    assert backend.datatype_allowed(StrictEnum)
    assert not backend.datatype_allowed(list)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_connect(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    backend = await make_backend(float, source, connect=False)
    assert backend is not None
    await backend.connect(1)
    backend.read_trl = ""
    with pytest.raises(RuntimeError) as exc:
        await backend.connect(1)
    assert "trl not set" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_connect_and_store_config(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    transport = await make_backend(float, source, connect=False)
    await transport._connect_and_store_config(source, 1)
    assert transport.trl_configs[source] is not None

    with pytest.raises(RuntimeError) as exc:
        await transport._connect_and_store_config("", 1)
    assert "trl not set" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_put(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnectedError) as exc:
        await transport.put(1.0)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    source = transport.source("", True)
    await transport.put(2.0)
    val = await transport.proxies[source].get_w_value()
    assert val == 2.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_datakey(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "limitedvalue")

    transport = TangoSignalBackend(float, trl, trl, device_proxy)

    with pytest.raises(NotConnectedError) as exc:
        await transport.get_datakey(transport.read_trl)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    datakey = await transport.get_datakey(trl)
    for key in ["source", "dtype", "shape", "limits", "precision", "units"]:
        assert key in datakey
    assert datakey["source"] == trl
    assert datakey["dtype"] == "number"
    assert datakey["shape"] == []
    for key in ["alarm", "control", "rds", "warning"]:
        assert key in datakey["limits"]
    limits = datakey["limits"]
    assert limits["alarm"] == {"high": 5.0, "low": 1.0}
    assert limits["control"] == {"high": 6.0, "low": 0.0}
    assert limits["rds"] == {"time_difference": 1.0, "value_difference": 1.0}
    assert limits["warning"] == {"high": 4.0, "low": 2.0}


# --------------------------------------------------------------------


@pytest.mark.asyncio
async def test_tango_transport_get_datakey_enum(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "test_enum")

    class TestEnumType(StrictEnum):
        A = "A"
        B = "B"

    transport = TangoSignalBackend(TestEnumType, trl, trl, device_proxy)
    await transport.connect(1)
    datakey = await transport.get_datakey(trl)
    assert "choices" in datakey
    assert datakey["choices"] == ["A", "B"]


# ---------------------------------------------------------------------


@pytest.mark.asyncio
async def test_tango_transport_get_reading(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnectedError) as exc:
        await transport.put(1.0)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    reading = await transport.get_reading()
    assert reading["value"] == 2.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_value(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnectedError) as exc:
        await transport.put(1.0)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    value = await transport.get_value()
    assert value == 2.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_setpoint(tango_test_device):
    source = get_full_attr_trl(tango_test_device, "floatvalue")
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnectedError) as exc:
        await transport.put(1.0)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    new_setpoint = 2.0
    await transport.put(new_setpoint)
    setpoint = await transport.get_setpoint()
    assert setpoint == new_setpoint


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_read_and_write_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    # Must use a FQTRL, at least on windows.
    read_trl = get_full_attr_trl(tango_test_device, "readback")
    write_trl = get_full_attr_trl(tango_test_device, "setpoint")

    # Test with existing proxy
    transport = TangoSignalBackend(float, read_trl, write_trl, device_proxy)
    await transport.connect(1)
    reading = await transport.get_reading()
    initial_value = reading["value"]
    new_value = initial_value + 1.0
    await transport.put(new_value)
    updated_value = await transport.get_value()
    assert updated_value == new_value

    # Without pre-existing proxy
    transport = TangoSignalBackend(float, read_trl, write_trl, None)
    await transport.connect(1)
    reading = await transport.get_reading()
    initial_value = reading["value"]
    new_value = initial_value + 1.0
    await transport.put(new_value)
    updated_value = await transport.get_value()
    assert updated_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_read_only_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    read_trl = get_full_attr_trl(tango_test_device, "readonly")

    # Test with existing proxy
    transport = TangoSignalBackend(int, read_trl, read_trl, device_proxy)
    await transport.connect(1)
    with pytest.raises(RuntimeError) as exc:
        await transport.put(1)
    assert "is not writable" in str(exc.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(14)
async def test_tango_transport_nonexistent_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    nonexistent_trl = get_full_attr_trl(tango_test_device, "nonexistent")

    # Test with existing proxy
    transport = TangoSignalBackend(int, nonexistent_trl, nonexistent_trl, device_proxy)
    with pytest.raises(RuntimeError) as exc:
        await transport.connect(1)
    assert "cannot be found" in str(exc.value)

    # Without pre-existing proxy
    transport = TangoSignalBackend(int, nonexistent_trl, nonexistent_trl, None)
    with pytest.raises(RuntimeError) as exc:
        await transport.connect(1)
    assert "cannot be found" in str(exc.value)


# ----------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(10)
async def test_type_mismatch_justvalue(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "justvalue")

    transport = TangoSignalBackend(float, trl, trl, device_proxy)
    with pytest.raises(TypeError) as exc:
        await transport.connect(1)
    val = str(exc.value)
    assert "has type" in val
    assert "int" in val
    assert "float" in val


@pytest.mark.asyncio
@pytest.mark.timeout(10)
async def test_type_mismatch_array(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "array")

    transport = TangoSignalBackend(npt.NDArray[int], trl, trl, device_proxy)
    with pytest.raises(TypeError) as exc:
        await transport.connect(1)
    val = str(exc.value)
    assert "has type numpy.ndarray" in val
    assert "numpy.dtype" in val
    assert "float" in val
    assert "int" in val


@pytest.mark.asyncio
@pytest.mark.timeout(10)
async def test_type_mismatch_sequence(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "sequence")

    transport = TangoSignalBackend(Sequence[int], trl, trl, device_proxy)
    with pytest.raises(TypeError) as exc:
        await transport.connect(1)
    val = str(exc.value)
    assert "has type" in val
    assert "str" in val
    assert "Sequence" in val
    assert "int" in val


@pytest.mark.asyncio
@pytest.mark.timeout(10)
async def test_type_mismatch_longstringarray(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = get_full_attr_trl(tango_test_device, "get_longstringarray")

    transport = TangoSignalBackend(TangoDoubleStringTable, trl, trl, device_proxy)
    with pytest.raises(TypeError) as exc:
        await transport.connect(1)
    val = str(exc.value)
    assert "has type" in val
    assert "TangoLongStringTable" in val
    assert "TangoDoubleStringTable" in val


@pytest.mark.asyncio
async def test_attribute_subscribe_event_fail(tango_test_device, caplog):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "nonexistent")
    attr_proxy.support_events = True

    def callback(reading, value):
        pass

    with caplog.at_level(logging.DEBUG, logger="ophyd_async"):
        attr_proxy.subscribe_callback(callback)
        await asyncio.sleep(0.1)

    assert any(
        "Subscribe to event failed" in record.message for record in caplog.records
    )



from enum import Enum

import pytest

from ophyd_async.core import StrictEnum, SupersetEnum
from ophyd_async.epics.core._util import get_supported_values  # noqa: PLC2701


def test_given_a_non_enum_passed_to_get_supported_enum_then_raises():
    with pytest.raises(TypeError):
        get_supported_values("", int, ("test",))


def test_given_an_enum_but_not_str_passed_to_get_supported_enum_then_raises():
    class MyEnum(Enum):
        TEST = "test"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test",))


def test_given_pv_has_choices_not_in_supplied_enum_then_raises():
    class MyEnum(StrictEnum):
        TEST = "test"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test", "unexpected_choice"))


def test_given_supplied_enum_has_choices_not_in_pv_then_raises():
    class MyEnum(StrictEnum):
        TEST = "test"
        OTHER = "unexpected_choice"

    with pytest.raises(TypeError):
        get_supported_values("", MyEnum, ("test",))


def test_given_a_supplied_enum_that_matches_the_pv_choices_then_enum_type_is_returned():
    class MyEnum(StrictEnum):
        TEST_1 = "test_1"
        TEST_2 = "test_2"

    supported_vals = get_supported_values("", MyEnum, ("test_1", "test_2"))
    assert len(supported_vals) == 2
    assert "test_1" in supported_vals
    assert "test_2" in supported_vals


def test_given_supersetenum_that_partial_matches_are_valid():
    class MyEnum(SupersetEnum):
        TEST_1 = "test_1"
        TEST_2 = "test_2"

    supported_vals = get_supported_values("", MyEnum, ("test_1",))

    assert "test_1" in supported_vals
    assert supported_vals.get("test_2") is None


class MySupersetEnum(SupersetEnum):
    TEST_1 = "test_1"
    TEST_2 = "test_2"


SUPERSETENUM_ERROR_MESSAGE = "to be a superset of them."


def test_given_supersetenum_that_all_values_plus_extra_values_are_invalid():
    with pytest.raises(TypeError, match=SUPERSETENUM_ERROR_MESSAGE):
        get_supported_values(
            "",
            MySupersetEnum,
            (MySupersetEnum.TEST_1, MySupersetEnum.TEST_2, "extra_1"),
        )


def test_given_supersetenum_that_partial_values_plus_extra_values_are_invalid():
    with pytest.raises(TypeError, match=SUPERSETENUM_ERROR_MESSAGE):
        get_supported_values("", MySupersetEnum, (MySupersetEnum.TEST_1, "extra_1"))


def test_given_supersetenum_that_no_matches_is_invalid():
    with pytest.raises(TypeError, match=SUPERSETENUM_ERROR_MESSAGE):
        get_supported_values("", MySupersetEnum, ("no_match_1", "no_match_2"))
